from PyQt5 import QtCore, QtWidgets, QtGui

from ..structures.structureddataframe import StructuredDataFrame


class PandasModel(QtCore.QAbstractTableModel):
    """
    Class to populate a table view with a pandas dataframe
    """

    def __init__(self, data, parent=None):
        QtCore.QAbstractTableModel.__init__(self, parent)
        self._data = data

    def rowCount(self, parent=None):
        return len(self._data.values)

    def columnCount(self, parent=None):
        return self._data.columns.size

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if index.isValid():
            if role == QtCore.Qt.DisplayRole:
                return str(self._data.values[index.row()][index.column()])
        return None

    def headerData(self, col, orientation, role):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return str(self._data.columns[col])
        return None


def viewDataFrame(df: StructuredDataFrame):
    """
    Construct a TableView Dialog to view DataFrameContents)

    Parameters
    ----------
    df : StructuredDataFrame
    """

    # TODO: Implement Copy Function

    dialog = QtWidgets.QDialog(None)
    layout = QtWidgets.QVBoxLayout(dialog)
    layout.setContentsMargins(0, 0, 0, 0)
    view = QtWidgets.QTableView(None)
    view.setModel(PandasModel(df))
    layout.addWidget(view)
    dialog.exec_()



import numpy as np
from collections import OrderedDict

from radie.structures import StructuredDataFrame, register_data_structures

CuKa1 = 1.5405980  # angstroms
CuKa2 = 1.5444260  # angstroms
CuKa = (2 * CuKa1 + CuKa2) / 3  # scaled avg of CuKa1 and 2, =1.541874
CuKb1 = 1.392250  # angstroms


def calc_Q(twotheta, wavelength):
    """
    Calculate the length of the reciprocal lattice vector given a diffraction angle and wavelength

    Parameters
    ----------
    twotheta : np.ndarray, float
        diffraction angle in degrees
    wavelength : float
        radiation wavelength in angstroms

    Returns
    -------
    Q : np.ndarray, float
        reciprocal lattice vector length

    """
    return 4 * np.pi / wavelength * np.sin(np.deg2rad(twotheta / 2))


def calc_twotheta(Q, wavelength):
    """
    Calcuate twotheta at a given wavelength from the length of the reciprocal lattice vector `Q`

    Parameters
    ----------
    Q : np.ndarray, float
        the length of the reciprocal lattice vector
    wavelength : float
        the desired wavelength in angstroms

    Returns
    -------
    twotheta : np.ndarray, float
        diffraction angle in degrees

    """
    return np.rad2deg(2 * np.arcsin(Q * wavelength / (2 * np.pi)))


def convert_wavelength(wl1, wl2, twotheta1):
    """
    Convert twotheta values to those for another wavelength

    Parameters
    ----------
    wl1 : float
        the original wavelength in angstroms
    wl2 : float
        the new wavelength in angstroms
    twotheta1 : np.ndarray, float
        the original twotheta values in degrees

    Returns
    -------
    twotheta2 : np.ndarray, float
        the new twotheta values in degrees

    """
    rad1 = np.deg2rad(twotheta1)
    rad2 = 2 * np.arcsin(wl2 / wl1 * np.sin(rad1 / 2))
    return np.rad2deg(rad2)


def calc_d_spacing(x, wavelength=None):
    """
    Calculate the d spacing given the x values, where x is either twotheta or the reciprocal lattice vector length Q

    Parameters
    ----------
    x : np.ndarray, float
        Q or twotheta.  if wavelength is provided, x is twotheta in degrees, otherwise Q
    wavelength : float, optional
        the wavelength

    Returns
    -------
    d_spacing : np.ndarray, float
        the lattice spacing in Angstroms

    """

    if wavelength is None:
        d_spacing = np.pi / x
    else:
        d_spacing = wavelength / (2 * np.sin(np.deg2rad(x) / 2))

    return d_spacing


class PowderDiffraction(StructuredDataFrame):
    """
    StructuredDataFrame for holding Powder diffraction data, valid for any bragg type diffraction measurements:
    Benchtop XRD, Synchrotron XRD, Neutron Diffraction
    """

    label = "Powder Diffraction"

    _required_metadata = StructuredDataFrame._required_metadata.copy()
    _required_metadata.update({
        "wavelength": CuKa,
        "source": "",
    })

    _x = "twotheta"
    _y = "intensity"

    _required_columns = OrderedDict((
        ("twotheta", float),
        ("intensity", float),
    ))
    _column_properties = ["Q", "d_spacing"]

    @property
    def Q(self):
        """return Q, the length of the reciprocal lattice vector"""
        if self.metadata["wavelength"] is None:
            raise ValueError("Must specify a wavelength to determine Q")
        srs = calc_Q(self.twotheta, self.metadata["wavelength"])
        srs.name = "Q"
        return srs

    def twotheta_at_wavelength(self, wavelength):
        return convert_wavelength(self.metadata["wavelength"], wavelength, self.twotheta)

    @property
    def d_spacing(self):
        """return d-spacing calculated from n=1 in the Bragg equation n * lambda = 2 * d sin(theta)"""
        if self.metadata["wavelength"] is None:
            raise ValueError("Must specify a wavelength to determine d-spacing")
        srs = calc_d_spacing(self.twotheta, self.metadata["wavelength"])
        srs.name = "d_spacing"
        return srs

register_data_structures(PowderDiffraction)



import typing
from collections import OrderedDict


class IndexableDict(OrderedDict):
    """
    A class that defines an ordered dictionary where the items are indexable with common slicing operations as well
    as by the dict key.  We can also retrieve the 'row' of the item as an integer by providing the key, or the value,
    in which case row will return the first instance of value
    """

    def __getitem__(self, key):
        if type(key) is slice:
            return tuple(self.values())[key]

        # can only use this functionality with slice because ints are valid dict keys, leaving here until I'm sure
        # all such uses are gone
        # elif type(key) is int:
        #     try:
        #         key = tuple(self)[key]
        #     except ValueError:
        #         raise ValueError("index {:} out of range".format(key))

        return super(IndexableDict, self).__getitem__(key)

    def getKey(self, index):
        """
        return the dict key at the given index

        Parameters
        ----------
        index : int

        Returns
        -------
        typing.Any
        """
        return tuple(self)[index]

    def getValue(self, index):
        """
        return the value at the given index

        Parameters
        ----------
        index : int

        Returns
        -------
        typing.Any
        """
        return tuple(self.values())[index]

    def getKeyAndValue(self, row):
        """
        return the ith key, value pair at row i

        Parameters
        ----------
        row : int

        Returns
        -------
        key : typing.Hashable
        value : typing.Any
        """
        key = self.getKey(row)
        return key, self[key]

    def rowCount(self):
        """
        QAbstractIndex convenience function

        Returns
        -------
        int
        """
        return len(self)

    def getKeyRow(self, key):
        """
        return the position of key in the dict

        Parameters
        ----------
        key : typing.Hashable

        Returns
        -------
        int
        """
        return tuple(self).index(key)

    def getValueRow(self, value):
        """
        return the position of the first occurence of value in the dict

        Parameters
        ----------
        value : typing.Any

        Returns
        -------
        int
        """
        return tuple(self.values()).index(value)



"""package for QtWidgets used to visualize StructuredDataFrame objects"""

import warnings
import typing
from collections import OrderedDict

from . import base, xyscatter, histogram, series

__all__ = []
visualizations = OrderedDict()


def register_visualizations(*sub_classes):
    """
    Parameters
    ----------
    sub_classes : typing.Type(base.Visualization)
        base.Visualization sub-classes passed in as arguments

    """
    for cls in sub_classes:
        if cls.__name__ in visualizations.keys():
            warnings.warn("overwriting visualization class {:}".format(cls.__name__))
        visualizations[cls.__name__] = cls

        if not cls.__name__ in __all__:
            globals()[cls.__name__] = cls  # put the class into this scope
            __all__.append(cls.__name__)


register_visualizations(
    xyscatter.XYScatter,
    series.SeriesVisualization,
    histogram.Histogram,
)



"""Module for the Visualization of particle size distributions
Log plot"""

import os

import pyqtgraph as pg

from radie.qt import colors, uchars
from radie.qt import plotlist
from radie.qt.visualizations import base, register_visualizations
from radie.plugins.structures import psd
from radie.plugins.visualizations.ui.psd import Ui_PSD


class DFItem(plotlist.DFItem):
    """an item class with pyqtgraph xy curve handles of type PlotDataItem"""
    def __init__(self, ref, item_list, name=None):
        super(DFItem, self).__init__(ref, item_list, name)
        self.plotDataItem = pg.PlotDataItem()
        #self.plotDataItem.setLogMode(x=True) #?
        self.color = None
        self.plotDataItem.setData(
            x=self.x_data(),
            y=self.y_data(),
            name=self.text
        )

    def setText(self, value):
        self.text = value
        self.plotDataItem.setData(name=value)
        self.plotDataItem.updateItems()


class VisPSD(base.Visualization, Ui_PSD):
    """A Visualization to compare particle size distributions"""

    name = "Particle Size Distribution"
    description = "A Visualization to compare particle size distributions"
    _icon_image = os.path.join(os.path.dirname(__file__), "icons/particlesizedistribution.svg")

    supportedClasses = (
        psd.PSD,
    )

    X_VALUES = [
        ["Diameter ({}m)".format(uchars.mu), None],
    ]

    def __init__(self, parent=None):
        super(VisPSD, self).__init__(parent)
        self.setupUi(self)
        self.splitter.setStretchFactor(0, 0)
        self.splitter.setStretchFactor(1, 100)

        self._colors = colors.colors()
        self.xmin = None
        self.xmax = None
        self.ymin = None
        self.ymax = None
        self._xFunc = None

        # setup class-types, and plotting-curve types for the list
        self.listView_datasets.supportedClasses = self.supportedClasses
        self.listView_datasets.setItemClass(DFItem)

#        for xVal in self.X_VALUES:
#            self.comboBox_xValues.addItem(xVal[0])
#        self.doubleSpinBox_wavelength.setValue(powderdiffraction.CuKa)

        self.plotWidget.plotItem.setLabel("left", "Frequency")
        self.plotWidget.plotItem.setLabel("bottom", "Diameter ({}m)".format(uchars.mu))
        self.plotWidget.plotItem.setLogMode(x=True)


        # --- signal connections --- #
        self.listView_datasets.model().itemsAdded.connect(self.addCurves)
        self.listView_datasets.model().itemTextUpdated.connect(self.plotWidget.plotItem.resetLegend)
        self.listView_datasets.model().itemToggled.connect(self.itemToggled)
        self.listView_datasets.model().rowsMoved.connect(self.processNewLayout)
        self.listView_datasets.model().itemsDeleted.connect(self.processNewLayout)
        # self.listView_datasets.model().rowsRemoved.connect(self.processNewLayout)
#        self.doubleSpinBox_wavelength.editingFinished.connect(self.determineXBounds)
#        self.doubleSpinBox_wavelength.editingFinished.connect(self.setXDataFunction)
#        self.doubleSpinBox_wavelength.editingFinished.connect(self.recalculateXValues)
#        self.checkBox_normalize.stateChanged.connect(self.recalculateYValues)
#        self.checkBox_xStagger.stateChanged.connect(self.recalculateXValues)
#        self.checkBox_xStagger.stateChanged.connect(self.doubleSpinBox_xStagger.setEnabled)
#        self.checkBox_yStagger.stateChanged.connect(self.recalculateYValues)
#        self.checkBox_yStagger.stateChanged.connect(self.doubleSpinBox_yStagger.setEnabled)
#        self.doubleSpinBox_yStagger.valueChanged.connect(self.recalculateYValues)
#        self.doubleSpinBox_xStagger.valueChanged.connect(self.recalculateXValues)
#        self.comboBox_xValues.currentIndexChanged.connect(self.xSelectionChanged)
        # --- end signal connections --- #

 #       self.doubleSpinBox_yStagger.setEnabled(False)
 #       self.doubleSpinBox_xStagger.setEnabled(False)
 #       self.setXDataFunction()

    def itemToggled(self, item):
        """
        process checking/unchecking of a StructuredDataFrame in the plot

        Parameters
        ----------
        item : DFItem
        """
        if item.checkState:
            item.plotDataItem.setData(
                x=self.xValues(item),
                y=self.yValues(item),
                name=item.text
            )
        else:
            item.plotDataItem.setData(name=None)
            item.plotDataItem.clear()
        self.plotWidget.plotItem.resetLegend()

    def processNewLayout(self):
        self.plotWidget.plotItem.clear()
        self.plotWidget.plotItem.resetLegend()
        self.recalculateYValues()
        for item in self.listView_datasets.iterItems():
            self.plotWidget.addItem(item.plotDataItem)

#    def xSelectionChanged(self):
#        xIndex = self.comboBox_xValues.currentIndex()
#        self.setXDataFunction()
#        self.determineXBounds()
#
#        self.doubleSpinBox_wavelength.setEnabled(xIndex not in (1, 2))
#
#        if xIndex == 1:
#            xLabel = "Q"
#        elif xIndex == 2:
#            xLabel = "d-space ({:s})".format(uchars.angstrom)
#        else:
#            xLabel = "two-theta (deg)"
#
#        self.plotWidget.plotItem.setLabel("bottom", xLabel)
#
#        if xIndex > 2:
#            wavelength = self.X_VALUES[xIndex][1]
#            self.doubleSpinBox_wavelength.setValue(wavelength)
#
#        self.recalculateXValues()

    def determineXBounds(self):
        # first, where are the xValues coming from
        items = self.listView_datasets.iterItems()
        item1 = next(items)
        x = self.xValues(item1)
        self.xmax = x.max()
        self.xmin = x.min()

        for item in items:
            x = self.xValues(item)
            df_xmax = x.max()
            if df_xmax > self.xmax:
                self.xmax = df_xmax
            df_xmin = x.min()
            if df_xmin > self.xmin:
                self.xmin = df_xmin

    def determineYBounds(self):
        items = self.listView_datasets.iterItems()
        item1 = next(items)
        y = self.yValues(item1)
        self.ymax = y.max()
        self.ymin = y.min()

        for item in items:
            y = self.yValues(item)
            df_ymax = y.max()
            if df_ymax > self.ymax:
                self.ymax = df_ymax
            df_ymin = y.min()
            if df_ymin > self.ymin:
                self.ymin = df_ymin

    def determineXYBounds(self):
        self.determineXBounds()
        self.determineYBounds()

    def recalculateXValues(self):
        for item in self.listView_datasets.iterItems():  # type: DFItem
            if item.checkState:
                item.plotDataItem.setData(
                    x=self.xValues(item),
                    y=item.plotDataItem.yData
                )

    def recalculateYValues(self):
        for item in self.listView_datasets.iterItems():  # type: DFItem
            if item.isChecked():
                item.plotDataItem.setData(
                    x=item.plotDataItem.xData,
                    y=self.yValues(item)
                )


    def xData(self, df):
        """

        Parameters
        ----------
        df : structures.PSD

        Returns
        -------
        xVals : np.ndarray
            the data to be used for x-axis plotting

        """
        return df.diameter.values

    def yData(self, df):
        """

        Parameters
        ----------
        df : structures.PSD

        Returns
        -------
        yVals : np.ndarray
            the data to be used for y-axis plotting
        """
        return df.frequency.values

    def xValues(self, item):
        """

        Parameters
        ----------
        item : DFItem

        Returns
        -------
        xVals : np.ndarray

        """
        xVals = self.xData(item.df)  # get the appropriate xValues, and scale them accordingly
        return xVals

    def yValues(self, item):
        """

        Parameters
        ----------
        item : DFItem

        Returns
        -------
        yVals : np.ndarray

        """
        yVals = self.yData(item.df)
        return yVals

    @property
    def nextColor(self):
        return next(self._colors)

    def resetColors(self):
        self._colors = colors.colors()

    def addCurves(self, listItems):
        """
        Parameters
        ----------
        listItems : list of DFItem
        """
        if not listItems:
            return

        self.determineXYBounds()

        for item in listItems:  # type: DFItem
            item.plotDataItem.setData(
                x=self.xValues(item),
                y=self.yValues(item),
                pen=self.nextColor
            )
            self.plotWidget.addItem(item.plotDataItem)


register_visualizations(VisPSD)



import struct
import numpy as np
from datetime import datetime

from radie import exceptions
from radie.loaders import Loader, register_loaders
from radie.plugins.structures.powderdiffraction import PowderDiffraction


def raw_check_version(f):
    """

    Parameters
    ----------
    f : file object

    Returns
    -------
    version : string or None
    """
    f.seek(0)
    head = read_str(f, 4)
    version = None

    if head == "RAW ":
        version = "ver. 1"
    elif head == "RAW2":
        version = "ver. 2"
    elif head == "RAW1" and read_str(f, 3) == ".01":
        version = "ver. 3"

    return version


def load_raw(fname, name=None):
    """
    .raw file output from Bruker XRD.  Tested with files from Bruker D8
    Ported from xylib\bruker_raw.cpp

    If multiple ranges exist in the RAW file a list of PowderDiffraction StructuredDataFrame
    objects are returned. If only one a single object is returned

    Parameters
    ----------
    fname : str
        filename
    name : str
        measurement identifier

    Returns
    -------
    df_xrd : single PowderDiffraction or list of them
        PowderDiffraction StructuredDataFrame based on XRD data

    """

    with open(fname, 'rb') as f:
        version = raw_check_version(f)
        if version == 'ver. 3':
            dfs = load_raw_version1_01(f, name=name)
        else:
            raise exceptions.IncorrectFileType

    if isinstance(dfs, PowderDiffraction) or type(dfs) in (list, tuple):
        if len(dfs) == 0:
            raise IOError('Unable to read scan from file')
        elif len(dfs) == 1:
            return dfs[0]
        else:
            return dfs
    else:
        raise TypeError("Unrecognized StructuredDataFrame type: {:}".format(type(dfs)))


def read_int(f):
    return struct.unpack('<i', f.read(4))[0]


def read_dbl(f):
    return struct.unpack('<d', f.read(8))[0]


def read_flt(f):
    return struct.unpack('<f', f.read(4))[0]


def read_str(f, len):
    return f.read(len).decode('utf-8').rstrip('\x00')


def load_raw_version1_01(f, name=None):
    """
    Read a file object pointing to a Bruker RAW file of version 1.01
    Since RAW files can contain more than one scan (or range) a
    list of PowderDiffraction dataframes are returned

    Parameters
    ----------
    f : file object

    Returns
    -------
    dfs : list of PowderDiffraction objects
    """
    meta = {}
    meta["format version"] = "3"

    f.seek(8)  # jump to after version line
    file_status = read_int(f)  # address 8

    if file_status == 1:
        meta["file status"] = "done"
    elif file_status == 2:
        meta["file status"] = "active"
    elif file_status == 3:
        meta["file status"] = "aborted"
    elif file_status == 4:
        meta["file status"] = "interrupted"

    range_cnt = read_int(f)  # address 12

    meta["MEASURE_DATE"] = read_str(f, 10)  # address 16
    meta["MEASURE_TIME"] = read_str(f, 10)  # address 26
    meta["USER"] = read_str(f, 72)  # address 36
    meta["SITE"] = read_str(f, 218)  # address 108
    meta["SAMPLE_ID"] = read_str(f, 60)  # address 326
    meta["COMMENT"] = read_str(f, 160)  # address 386
    f.read(2)  # apparently there is a bug in docs, 386+160 != 548
    f.read(4)  # goniometer code # address 548
    f.read(4)  # goniometer stage code # address 552
    f.read(4)  # sample loader code # address 556
    f.read(4)  # goniometer controller code # address 560
    f.read(4)  # (R4) goniometer radius # address 564
    f.read(4)  # (R4) fixed divergence...# address 568
    f.read(4)  # (R4) fixed sample slit...# address 572
    f.read(4)  # primary Soller slit # address 576
    f.read(4)  # primary monochromator # address 580
    f.read(4)  # (R4) fixed antiscatter...# address 584
    f.read(4)  # (R4) fixed detector slit...# address 588
    f.read(4)  # secondary Soller slit # address 592
    f.read(4)  # fixed thin film attachment # address 596
    f.read(4)  # beta filter # address 600
    f.read(4)  # secondary monochromator # address 604
    meta["ANODE_MATERIAL"] = read_str(f, 4)  # address 608
    f.read(4)  # unused # address 612
    meta["ALPHA_AVERAGE"] = read_dbl(f)  # address 616
    meta["ALPHA1"] = read_dbl(f)  # address 624
    meta["ALPHA2"] = read_dbl(f)  # address 632
    meta["BETA"] = read_dbl(f)  # address 640
    meta["ALPHA_RATIO"] = read_dbl(f)  # address 648
    f.read(4)  # (C4) unit name # address 656
    f.read(4)  # (R4) intensity beta:a1 # address 660
    meta["measurement time"] = read_flt(f)  # address 664
    f.read(43)  # unused # address 668
    f.read(1)  # hardware dependency... # address 711
    # assert (f.tellg() == 712);

    # Expected meta
    # Convert to datetime
    timestamp = datetime.strptime('{} {}'.format(meta["MEASURE_DATE"], meta["MEASURE_TIME"]), '%m/%d/%y %H:%M:%S')
    meta['date'] = timestamp.isoformat()

    # range header
    dfs = []
    for cur_range in range(range_cnt):

        if name is None:
            meta["name"] = '{}-{}'.format(meta["SAMPLE_ID"], cur_range)
        else:
            meta["name"] = '{}-{}'.format(name, cur_range)

        # Take off the number of only one
        if range_cnt == 1:
            meta["name"] = meta["name"][:-2]

        # Block * blk = new Block;
        header_len = read_int(f)  # address 0
        assert header_len == 304

        steps = read_int(f)  # address 4
        meta["STEPS"] = steps
        start_theta = read_dbl(f)  # address 8
        meta["START_THETA"] = start_theta
        start_2theta = read_dbl(f)  # address 16
        meta["START_2THETA"] = start_2theta

        f.read(8)  # Chi drive start # address 24
        f.read(8)  # Phi drive start # address 32
        f.read(8)  # x drive start # address 40
        f.read(8)  # y drive start # address 48
        f.read(8)  # z drive start # address 56
        f.read(8)  # address 64
        f.read(6)  # address 72
        f.read(2)  # unused # address 78
        f.read(8)  # (R8) variable antiscat.# address 80
        f.read(6)  # address 88
        f.read(2)  # unused # address 94
        f.read(4)  # detector code # address 96
        meta["HIGH_VOLTAGE"] = read_flt(f)  # address 100
        meta["AMPLIFIER_GAIN"] = read_flt(f)  # 104
        meta["DISCRIMINATOR_1_LOWER_LEVEL"] = read_flt(f)  # 108
        f.read(4)  # address 112
        f.read(4)  # address 116
        f.read(8)  # address 120
        f.read(4)  # address 128
        f.read(4)  # address 132
        f.read(5)  # address 136
        f.read(3)  # unused # address 141
        f.read(8)  # address 144
        f.read(8)  # address 152
        f.read(8)  # address 160
        f.read(4)  # address 168
        f.read(4)  # unused # address 172
        step_size = read_dbl(f)  # address 176
        meta["STEP_SIZE"] = step_size
        f.read(8)  # address 184
        meta["TIME_PER_STEP"] = read_flt(f)  # 192
        f.read(4)  # address 196
        f.read(4)  # address 200
        f.read(4)  # address 204
        meta["ROTATION_SPEED [rpm]"] = read_flt(f)  # 208
        f.read(4)  # address 212
        f.read(4)  # address 216
        f.read(4)  # address 220
        meta["GENERATOR_VOLTAGE"] = read_int(f)  # 224
        meta["GENERATOR_CURRENT"] = read_int(f)  # 228
        f.read(4)  # address 232
        f.read(4)  # unused # address 236
        meta["USED_LAMBDA"] = read_dbl(f)  # 240
        f.read(4)  # address 248
        f.read(4)  # address 252
        supplementary_headers_size = read_int(f)  # address 256
        f.read(4)  # address 260
        f.read(4)  # address 264
        f.read(4)  # unused # address 268
        f.read(8)  # address 272
        f.read(24)  # unused # address 280
        # assert (f.tellg() == 712 + (cur_range + 1) * header_len);

        if supplementary_headers_size > 0:
            f.read(supplementary_headers_size)

        # StepColumn * xcol = new StepColumn(start_2theta, step_size);
        # blk->add_column(xcol);
        # VecColumn * ycol = new VecColumn;

        xcol = np.array(range(steps)) * step_size + start_2theta
        ycol = []
        for i in range(steps):
            ycol.append(read_flt(f))
        ycol = np.array(ycol)

        data = np.c_[xcol, ycol]
        df_meta = meta.copy()
        if meta["ANODE_MATERIAL"].startswith('Cu'):
            source = 'CuKa'
        else:
            raise ValueError("Unimplemented Anode Material {}".format(meta["ANODE_MATERIAL"]))
        df_xrd = PowderDiffraction(data=data,
                                   columns=['twotheta', 'intensity'],
                                   wavelength=df_meta['ALPHA_AVERAGE'],
                                   source=source,
                                   xunit="deg",
                                   yunit="counts",
                                   name=df_meta["name"],
                                   date=df_meta["date"],
                                   metadata=df_meta)

        dfs.append(df_xrd)

    return dfs


bruker_raw_loader = Loader(load_raw, PowderDiffraction, [".raw"], "Bruker RAW XRD")

register_loaders(
    bruker_raw_loader,
)



"""convenience testing package to generate example DataFrames from example data"""
import os

from ..loaders import vsm_lakeshore, powderdiffraction_rigaku

this_dir = os.path.dirname(__file__)


def example_powderdiffraction():
    """
    return an example `PowderDiffraction` generated by an example data file

    Returns
    -------
    df : PowderDiffraction

    """
    example_ras_file = os.path.join(this_dir, "data/xrd.ras")
    df = powderdiffraction_rigaku.load_ras(example_ras_file)
    return df


def example_vsm():
    """
    return an example `VSM` generated by an example data file

    Returns
    -------
    df : VSM

    """
    example_file = os.path.join(this_dir, "data/vsm.txt")
    df = vsm_lakeshore.load_ideavsm_txt(example_file)
    return df



class RadieException(Exception):
    """Define a base class for Radie Exceptions, and subclass all specific exceptions from this"""
    pass


class LoaderException(RadieException):
    """Generic exception for expected bad things happening inside loader functions"""
    pass


class LoaderNotFound(LoaderException):
    """Raise when no loader was found"""
    pass


class IncorrectFileType(LoaderException):
    """Raise this exception when a file is determined to be an incorrect type for the loader function"""
    pass



from collections import OrderedDict
from radie.structures import StructuredDataFrame, register_data_structures


class VSM(StructuredDataFrame):
    """
    A class for Vibrating Sample Magnetometer measurements.  Although these measurements can get quite complicated,
    measurements versus angle, the typical measurement is a measurement of applied field (in Gauss) versus magnetic
    moment of the sample (in emu)
    """

    label = "VSM"

    _required_metadata = StructuredDataFrame._required_metadata.copy()
    _required_metadata.update({
        "mass": 1,
        "density": 1
    })

    _x = "Field"
    _y = "Moment"

    _required_columns = OrderedDict((
        ("Field", float),
        ("Moment", float),
    ))
    _column_properties = []
    _loaders = []


register_data_structures(VSM)



"""
define a QAbstractItemModel for a an Ordered Dict of an OrderedDict of DataFrames, and also a
TreeView of the Items
"""

from PyQt5 import QtCore, QtWidgets

from .indexabledict import IndexableDict
from ..structures import StructuredDataFrame
from . import dataframeview


class DataStructureTree(IndexableDict):
    """The main StructuredDataFrame tree Construct

    This Class is the Root of the Tree of DataFrames presented in the GUI viewer.  Essentially, it is an Ordered
    Dict whose keys are the StructuredDataFrame classes, and whose values are the record of the available instances of those
    classes

    """
    def __init__(self):
        super(DataStructureTree, self).__init__()
        self.uuid = {}  # a place to track down DataFrames by uuid

    def __setitem__(self, key, value):
        if not issubclass(key, StructuredDataFrame):
            raise TypeError("key must be a subclass of StructuredDataFrame")
        if not isinstance(value, DataStructureNode):
            raise TypeError("DFStructureTree items must be of type DataStructureNode")
        super(DataStructureTree, self).__setitem__(key, value)

    def addDataFrame(self, df):
        """the main api for adding in new DataFrames

        Parameters
        ----------
        df : StructuredDataFrame

        Returns
        -------
        ref : DFReference

        """
        uuid = df.get_uuid()
        if uuid in self.uuid.keys():
            raise ValueError("df already exists")
        insert_row = -1

        cls = df.__class__
        node = self.get(cls, None)
        if node is None:
            node = self[cls] = DataStructureNode(cls, self)
            insert_row = len(self)
        ref = node[uuid] = DFReference(df, node)
        self.uuid[uuid] = ref

        return ref, node, insert_row


class DataStructureNode(IndexableDict):
    """A foldable node on the TreeView, one for each class of StructuredDataFrame

    A dict whose keys are a uuid string generated by the StructuredDataFrame instance, and the values are a reference to
    the StructuredDataFrame of the class specified by the cls attribute

    Attributes
    ----------
    tree : DataStructureTree
        A reference to the parent tree
    cls : typing.Type(StructuredDataFrame)
        The subclass of StructuredDataFrame that all members of this node belong to
    """
    def __init__(self, cls, tree):
        self.cls = cls  # type: StructuredDataFrame
        self.tree = tree  # type: DataStructureTree
        super(DataStructureNode, self).__init__()

    def parent(self):
        return self.tree

    def row(self):
        return self.tree.getKeyRow(self.cls)

    def pop(self, key):
        reference = super(DataStructureNode, self).pop(key)  # type: DFReference
        reference.dfDeleted.emit(reference)

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError("keys must be a uuid string")
        elif not isinstance(value, DFReference):
            raise TypeError("value must be of type {:}".format(DFReference.__name__))

        super(DataStructureNode, self).__setitem__(key, value)


class DFReference(QtCore.QObject):
    """A QObject which serves as a reference to a DataFrame instance, and can emit Qt Signals when needed

    Attributes
    ----------
    df : StructuredDataFrame
        The instance of a DataFrame that this reference refers to
    node : DataStructureNode
        a reference to the parent node of this reference
    row : int
        the location of this item in the node
    """
    dataChanged = QtCore.pyqtSignal()
    dfDeleted = QtCore.pyqtSignal(object)

    def __init__(self, df: StructuredDataFrame, node: DataStructureNode):
        if not isinstance(df, StructuredDataFrame):
            raise TypeError("df must be a StructuredDataFrame")
        self._df = df
        self.node = node
        super(DFReference, self).__init__()

    @property
    def df(self):
        return self._df

    def parent(self):
        return self.node

    def row(self):
        return self.node.getValueRow(self)

    @staticmethod
    def rowCount():
        return 0


class DFTreeModel(QtCore.QAbstractItemModel):
    headers = ["Name", "uuid"]

    def __init__(self, root, parent=None):
        """
        Parameters
        ----------
        root : DataStructureTree
        parent : QtWidgets.QWidget
        """
        super(DFTreeModel, self).__init__(parent)
        self.root = root
        for i, header in enumerate(self.headers):
            self.setHeaderData(i, QtCore.Qt.Horizontal, header)

    def index(self, row: int, column: int, parent: QtCore.QModelIndex = ...) -> QtCore.QModelIndex:
        if not self.hasIndex(row, column, parent):
            return QtCore.QModelIndex()

        if not parent.isValid():
            pointer = self.root  # type: DataStructureTree
        else:
            pointer = parent.internalPointer()  # type: DataStructureNode
        child = pointer.getValue(row)

        index = self.createIndex(row, column, child)
        return index

    def headerData(self, section: int, orientation: QtCore.Qt.Orientation, role: int = ...):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return self.headers[section]
        else:
            return None

    def structureNodeIndex(self, cls, create=True):
        rootIndex = QtCore.QModelIndex()
        if cls not in self.root.keys():
            if not create:
                raise IndexError

            row = len(self.root)
            self.beginInsertRows(rootIndex, row, row)
            self.root[cls] = DataStructureNode(cls, self.root)
            self.endInsertRows()
        else:
            row = self.root.getKeyRow(cls)
        nodeIndex = self.index(row, 0, rootIndex)
        return nodeIndex

    def deleteDataFrame(self, index: QtCore.QModelIndex):
        df = index.internalPointer()
        if not isinstance(df, DFReference):
            raise TypeError("index does not contain a StructuredDataFrame")
        row = index.row()
        self.beginRemoveRows(self.parent(index), row, row)
        node = df.parent()
        node.pop(df.df.uuid)
        self.endRemoveRows()

    def addDataFrame(self, df):
        """
        API for adding new dataframes after the model has been initiated

        Parameters
        ----------
        df : StructuredDataFrame

        """
        cls = df.__class__
        nodeIndex = self.structureNodeIndex(cls, create=True)
        node = nodeIndex.internalPointer()
        row = len(node)
        self.beginInsertRows(nodeIndex, row, row)
        self.root.addDataFrame(df)
        self.endInsertRows()
        return nodeIndex

    def parent(self, child: QtCore.QModelIndex) -> QtCore.QModelIndex:
        if not child.isValid():
            return QtCore.QModelIndex()

        item = child.internalPointer()
        if isinstance(item, DFReference):
            node = item.parent()
            row = node.row()
            return self.createIndex(row, 0, node)
        else:  # item is DataStructureNode
            return QtCore.QModelIndex()

    def rowCount(self, parent: QtCore.QModelIndex = ...) -> int:
        if parent.isValid():
            pointer = parent.internalPointer()
            return pointer.rowCount()
        else:
            return self.root.rowCount()

    def columnCount(self, parent: QtCore.QModelIndex = ...) -> int:
        return 2  # 1: class_name or FamilyMember.name | 2: FamilyMember.age

    def hasChildren(self, parent: QtCore.QModelIndex = ...) -> bool:
        item = parent.internalPointer()
        if isinstance(item, DFReference):
            return False
        else:
            return True

    def data(self, index: QtCore.QModelIndex, role: int = ...):
        if not index.isValid():
            return None
        if role in (QtCore.Qt.DisplayRole, QtCore.Qt.EditRole):
            item = index.internalPointer()
            if isinstance(item, DataStructureNode):
                if index.column() == 0:
                    return item.cls.__name__
            elif isinstance(item, DFReference):
                if index.column() == 0:
                    return item._df.metadata["name"]
                elif index.column() == 1:
                    return item._df.get_uuid()
                else:
                    return None
        else:  # Not display or Edit
            return None

    def setData(self, index: QtCore.QModelIndex, value, role: int = ...):
        item = index.internalPointer()
        col = index.column()
        if isinstance(item, DFReference):
            if col == 0:
                item._df.metadata["name"] = str(value)
            else:
                return False
            self.dataChanged.emit(index, index)
            return True
        return False

    def flags(self, index: QtCore.QModelIndex):
        if not index.isValid():
            return QtCore.Qt.NoItemFlags
        item = index.internalPointer()
        if isinstance(item, DFReference):
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEditable | \
                   QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled
        elif isinstance(item, DataStructureNode):
            return QtCore.Qt.ItemIsEnabled
        else:
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable


class DFTreeView(QtWidgets.QTreeView):
    def __init__(self, data=None, parent=None):
        super(DFTreeView, self).__init__(parent)
        if not isinstance(data, DataStructureTree):
            self.data = DataStructureTree()
        else:
            self.data = data
        model = DFTreeModel(self.data)
        self.setModel(model)
        rootIndex = QtCore.QModelIndex()
        for row in range(len(self.data)):
            self.setExpanded(model.index(row, 0, rootIndex), True)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.rightClicked)
        self.setSelectionMode(self.ExtendedSelection)
        self.setDragDropMode(self.DragDrop)

    def getSelectedDataFrameIndices(self):
        selected_df_indices = list()
        for index in self.selectionModel().selectedRows(0):  # type: QtCore.QModelIndex
            pointer = index.internalPointer()
            if isinstance(pointer, DFReference):
                selected_df_indices.append(index)
        return selected_df_indices

    def deleteSelectedDataFrames(self):
        for index in self.getSelectedDataFrameIndices():  # type: QtCore.QModelIndex
            self.model().deleteDataFrame(index)

    def viewSelectedDataFrame(self):
        row = self.selectionModel().selectedRows(0)[0]
        pointer = row.internalPointer()
        if isinstance(pointer, DFReference):
            dataframeview.viewDataFrame(pointer.df)

    def copySelectedDataFrame(self):
        row = self.selectionModel().selectedRows(0)[0]
        pointer = row.internalPointer()
        if isinstance(pointer, DFReference):
            pointer.df.to_clipboard(excel=True)

    def rightClicked(self, pos: QtCore.QPoint):
        menu = QtWidgets.QMenu()

        selected_rows = self.selectionModel().selectedRows(column=0)
        num_dataframes = len(selected_rows)

        if num_dataframes > 0:
            if num_dataframes == 1:
                deleteItem = QtWidgets.QAction("Delete StructuredDataFrame", menu)
                viewItem = QtWidgets.QAction("View StructuredDataFrame", menu)
                viewItem.triggered.connect(self.viewSelectedDataFrame)
                menu.addAction(viewItem)
                copyItem = QtWidgets.QAction("Copy StructuredDataFrame", menu)
                copyItem.triggered.connect(self.copySelectedDataFrame)
                menu.addAction(copyItem)
            else:
                deleteItem = QtWidgets.QAction("Delete DataFrames", menu)

            deleteItem.triggered.connect(self.deleteSelectedDataFrames)
            menu.addAction(deleteItem)

        menu.exec_(self.mapToGlobal(pos))  # QtWidgets.QAction

    def addDataFrame(self, df):
        model = self.model()  # type: DFTreeModel
        nodeIndex = model.addDataFrame(df)
        self.setExpanded(nodeIndex, True)
        self.resizeColumnToContents(0)


def test():
    """function to test DFTree from"""
    import sys
    app = QtWidgets.QApplication(sys.argv)
    from radie.plugins import examples
    vsm1 = examples.example_vsm()
    vsm2 = examples.example_vsm()
    pow1 = examples.example_powderdiffraction()
    pow2 = examples.example_powderdiffraction()

    data = DataStructureTree()
    for df in (vsm1, vsm2):
        data.addDataFrame(df)

    def _excepthook(e, v, t):
        sys.__excepthook__(e, v, t)

    sys.excepthook = _excepthook
    tree_view = DFTreeView(data)
    tree_view.show()
    pow2.name = "Second XRD"
    for df in (pow1, pow2):
        tree_view.addDataFrame(df)
    sys.exit(app.exec_())



"""a little module for handling colors and color generators"""

_mpl_hex_colors = ['1f77b4', 'ff7f0e', '2ca02c', 'd62728', '9467bd', '8c564b', 'e377c2', '7f7f7f', 'bcbd22', '17becf']


def colors(color_list=_mpl_hex_colors):
    """
    a simple generator to get the next color in the list

    Parameters
    ----------
    color_list : list
        a list of colors, usually hex strings, or other valid pyqgtraph Pen color specifiers

    Returns
    -------
    color : str
        returns whatever type of item in the list is

    """
    count = 0
    num = len(color_list)
    while True:
        yield color_list[count % num]
        count += 1



import ctypes
import os
import platform
import sys
import traceback
from os import path

from PyQt5 import QtWidgets, QtGui

from .classes import TextWarning, WarningMsgBox
from . import cfg

last_path = os.path.normpath(cfg.user_path)


def instantiate_app(sys_argv=[]):
    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication(sys_argv)
    cfg.set_dpi_scaling()
    return app


def excepthook(*args):
    return sys.__excepthook__(*args)


def reset_excepthook():
    sys.excepthook = excepthook


def popup_excepthook(type, value, tback):
    WarningMsgBox(traceback.format_exception(type, value, tback), "Uncaught Exception").exec_()


def set_popup_exceptions():
    sys.excepthook = popup_excepthook


def error_popup(error):
    if not type(error) is str:
        etype, value, tb = sys.exc_info()
        error = "\n".join(traceback.format_exception(etype, value, tb))
    WarningMsgBox(error).exec_()


def text_based_error(text):
    warning = TextWarning(text)
    warning.exec_()


def warnMissingFeature():
    msg = "Coming soon..."
    QtWidgets.QMessageBox.warning(None, "Feature Missing", msg, QtWidgets.QMessageBox.Ok)


def getOpenFileName_Global(caption, filter, start_path=None, **kwargs):
    global last_path
    if start_path is None:
        start_path = last_path
    fname = str(QtWidgets.QFileDialog.getOpenFileName(None, caption, start_path, filter, **kwargs)[0])
    if fname in ("", None):
        return ""
    last_path = path.dirname(fname)
    return fname


def getOpenFileNames_Global(caption, filter, start_path=None, **kwargs):
    global last_path
    if start_path is None:
        start_path = last_path
    fnames = QtWidgets.QFileDialog.getOpenFileNames(None, caption, start_path, filter, **kwargs)[0]
    fnames = [str(fname) for fname in fnames]
    if fnames in ("", None, []):
        return []
    last_path = path.dirname(fnames[0])
    return fnames


def getSaveFileName_Global(caption, filter, start_path=None, **kwargs):
    global last_path
    if start_path is None:
        start_path = last_path
    fname = str(QtWidgets.QFileDialog.getSaveFileName(None, caption, start_path, filter, **kwargs)[0])
    if fname in ("", None):
        return ""
    last_path = path.dirname(fname)
    return fname


def getDirName_Global(caption=None, start_path=None, **kwargs):
    global last_path
    if start_path is None:
        start_path = last_path
    dirname = str(QtWidgets.QFileDialog.getExistingDirectory(None, caption, start_path, **kwargs))
    if dirname in ("", None):
        return ""
    last_path = path.dirname(dirname)
    return dirname


def set_process_id(appid=None):
    """
    in windows, setting this parameter allows all instances to be grouped under the same taskbar icon, and allows
    us to set an icon that is different from whatever the python executable is using.
    :param appid: str, unicode
    :return:
    """
    if appid and type(appid) is str and platform.system() == "Windows":
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(appid)


def setup_style(style=cfg.preferred_style):
    """

    Parameters
    ----------
    style : str
        the Qt Style keyword matching the system style, default is fusion
    """
    available_styles = QtWidgets.QStyleFactory.keys()
    if style:
        if "QT_STYLE_OVERRIDE" in os.environ.keys():
            os.environ.pop("QT_STYLE_OVERRIDE")

        if style in available_styles:
            QtWidgets.QApplication.setStyle(style)
        else:
            for s in cfg.preferred_styles:
                if s in available_styles:
                    QtWidgets.QApplication.setStyle(s)

    elif platform.system() != "Windows" and os.environ["QT_API"] == "pyqt5":
        if "QT_STYLE_OVERRIDE" in os.environ.keys():
            os.environ.pop("QT_STYLE_OVERRIDE")
        if len(available_styles) == 2:
            # available styles are Windows, and Fusion
            # qt5-style-plugins are not installed, take action:
            for s in cfg.preferred_styles:
                if s in available_styles:
                    QtWidgets.QApplication.setStyle(s)


def icon(filename):
    """
    a convience function to get QIcons from the radie/qt/resources/icons directory

    Parameters
    ----------
    filename : str

    Returns
    -------
    QtGui.QIcon
    """
    return QtGui.QIcon(os.path.join(cfg.icon_path, filename))



"""plot a line of a single pandas series"""

import os

import numpy as np
from PyQt5 import QtCore, QtWidgets
import pyqtgraph as pg

from ...structures.structureddataframe import StructuredDataFrame
from .. import colors, cfg, dpi
from .. import plotwidget
from .. import plotlist
from .base import Visualization


class DFItem(plotlist.DFItem):
    """an item class with pyqtgraph xy curve handles of type PlotDataItem"""
    def __init__(self, ref, item_list, name=None):
        super(DFItem, self).__init__(ref, item_list, name)
        self.plotDataItem = pg.PlotDataItem()
        self.color = None
        self.values = None

    def setText(self, value):
        self.text = value
        self.plotDataItem.setData(name=value)
        self.plotDataItem.updateItems()

    def calculateValue(self):
        """calculate the histogram values, given the number of bins

        Parameters
        ----------
        bins : int
        density : bool

        """
        self.values = self.x_data()


class SeriesVisualization(Visualization):
    """A generic class providing the visualization of a column of data as evenly-spaced points"""

    name = "Series Line"
    description = "A generic visualization of Series of values"
    _icon_image = os.path.join(cfg.icon_path, "seriesvisualization.svg")

    def __init__(self, parent=None):
        super(SeriesVisualization, self).__init__(parent)

        # --- setupUi --- #
        self.resize(1000, 600)
        self.setWindowTitle("Series")
        self.vlay_main = QtWidgets.QVBoxLayout(self)
        self.vlay_main.setContentsMargins(3, 3, 3, 3)
        self.splitter = QtWidgets.QSplitter(self)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.dataListWidget = QtWidgets.QWidget(self.splitter)
        self.vlay_dataListWidget = QtWidgets.QVBoxLayout(self.dataListWidget)
        self.vlay_dataListWidget.setContentsMargins(0, 0, 0, 0)
        self.treeView_datasets = plotlist.DFSeriesListView(self.dataListWidget)
        self.vlay_dataListWidget.addWidget(self.treeView_datasets)
        self.hlay_parameters = QtWidgets.QHBoxLayout()
        self.vlay_dataListWidget.addLayout(self.hlay_parameters)
        self.plotWidget = plotwidget.PlotWidget(self.splitter)
        self.vlay_main.addWidget(self.splitter)

        self.splitter.setStretchFactor(0, 0)
        self.splitter.setStretchFactor(1, 1)
        self.splitter.setSizes(dpi.width_by_height(280, 720))
        # --- end setupUi --- #

        self.treeView_datasets.setItemClass(DFItem)
        self.treeView_datasets.supportedClasses = self.supportedClasses

        self.treeView_datasets.model().itemsAdded.connect(self.addCurves)
        self.treeView_datasets.model().itemAccessorChanged.connect(self.itemDataChanged)
        self.treeView_datasets.model().itemsDeleted.connect(self.processNewLayout)
        self.treeView_datasets.model().itemToggled.connect(self.itemToggled)
        self.treeView_datasets.model().itemTextUpdated.connect(self.plotWidget.plotItem.resetLegend)
        # self.checkBox_plotDensity.stateChanged.connect(self.recalculateValues)
        # self.spinBox_bins.editingFinished.connect(self.recalculateValues)

        self._colors = None
        self.resetColors()

    def nextColor(self):
        return next(self._colors)

    def resetColors(self):
        self._colors = colors.colors()

    def recalculateValues(self):
        for item in self.treeView_datasets.iterItems():
            item.calculateValue()
            item.plotDataItem.setData(x=np.arange(len(item.values)), y=item.values)

    def itemDataChanged(self, item: DFItem):
        if not item.isChecked():
            return
        item.calculateValue()
        item.plotDataItem.setData(x=np.arange(len(item.values)), y=item.values)

    def processNewLayout(self):
        self.plotWidget.plotItem.clear()
        self.plotWidget.plotItem.resetLegend()
        for item in self.treeView_datasets.iterItems():
            self.plotWidget.addItem(item.plotDataItem)

    def itemToggled(self, item):
        """
        process checking/unchecking of a StructuredDataFrame in the plot

        Parameters
        ----------
        item : DFItem
        """
        if item.checkState:
            item.plotDataItem.setData(x=np.arange(len(item.values)), y=item.values)
        else:
            item.plotDataItem.setData(name=None, stepMode=False)
            item.plotDataItem.clear()
        self.plotWidget.plotItem.resetLegend()

    def addCurves(self, items):
        """
        Parameters
        ----------
        items : list of DFItem
        """
        for item in items:  # type: DFItem
            item.color = self.nextColor()
            item.calculateValue()
            item.plotDataItem.setData(x=np.arange(len(item.values)), y=item.values, pen=item.color)
            self.plotWidget.plotItem.addItem(item.plotDataItem)


def test():
    import sys
    from ..masterdftree import DFReference
    from .. import cfg
    from .. import functions as fn
    app = fn.instantiate_app()
    fn.reset_excepthook()
    cfg.set_dpi_scaling()

    rng = np.random.RandomState(10)  # deterministic random data
    a = np.hstack((rng.normal(size=1000), rng.normal(loc=5, scale=2, size=1000)))
    b = np.hstack((rng.normal(size=1000, scale=2), rng.normal(loc=5, scale=1, size=1000)))
    data = np.vstack((a, b)).T
    df = StructuredDataFrame(data, columns=("a", "b"), name="Deterministic Random Data")
    ref1 = DFReference(df, None)

    plot = SeriesVisualization()
    plot.treeView_datasets.addDataFrames(ref1, ref1)
    plot.show()

    sys.exit(app.exec_())



"""define loader objects that return PowderDiffraction Data Structures"""
import re
import numpy as np
from radie.loaders import Loader, register_loaders

from radie.plugins.structures.powderdiffraction import PowderDiffraction, CuKa


def load_txt(fname):
    """
    .txt file output from Siemens D500 on custom 3M Software

    Parameters
    ----------
    fname : str
        filename

    Returns
    -------
    df_xrd : PowderDiffraction
        PowderDiffraction StructuredDataFrame based on XRD data

    """

    with open(fname, "r") as fid:
        lines = fid.readlines()

    lines = [line.strip() for line in lines]
    settings = lines.pop(0)
    mo = re.compile\
        (r'^"(?P<name>.*)",(?P<min>[0-9\.]+),(?P<max>[0-9\.]+),(?P<step>[0-9\.]+),(?P<count>[0-9\.]+).*$')
    meta = mo.match(settings).groupdict()

    for k in ['min', 'max', 'step', 'count']:
        meta[k] = float(meta[k])

    data = np.array([[float(x) for x in line.split()] for line in lines])
    df_xrd = PowderDiffraction(data=data,
                               columns=['twotheta', 'intensity'],
                               wavelength=CuKa,
                               source='CuKa',
                               xunit="deg",
                               yunit="counts",
                               **meta)

    return df_xrd


siemensD500_txt_loader = Loader(load_txt, PowderDiffraction, [".txt"], "Siemens D500")

register_loaders(
    siemensD500_txt_loader,
)



"""Module for the Visualizations of PowderDiffraction Datasets"""

import os

import pyqtgraph as pg

from radie.qt import colors, uchars
from radie.qt import plotlist
from radie.qt.visualizations import base, register_visualizations
from radie.plugins.structures import powderdiffraction
from radie.plugins.visualizations.ui.powderdiffraction import Ui_PowderDiffraction


class DFItem(plotlist.DFItem):
    """an item class with pyqtgraph xy curve handles of type PlotDataItem"""
    def __init__(self, ref, item_list, name=None):
        super(DFItem, self).__init__(ref, item_list, name)
        self.plotDataItem = pg.PlotDataItem()
        self.plotDataItem.setData(name=self.text)

    def setText(self, value):
        self.text = value
        self.plotDataItem.setData(name=value)


class VisPowderDiffraction(base.Visualization, Ui_PowderDiffraction):
    """Visualization to compare powder diffraction histograms

    Attributes
    ----------
    x_values : list
        a list of key, value pairs for the comboBox_xValues widget

    """
    name = "Powder Diffraction"
    description = "A Visualization to compare powder diffraction histograms"
    _icon_image = os.path.join(os.path.dirname(__file__), "icons/powderdiffraction.svg")

    supportedClasses = (
        powderdiffraction.PowderDiffraction,
    )

    x_values = [
        ["wavelength ({:})".format(uchars.angstrom), None],
        ["Q", None],
        ["d-spacing", None],
        ["CuKa", powderdiffraction.CuKa]
    ]

    def __init__(self, parent=None):
        super(VisPowderDiffraction, self).__init__(parent)
        self.setupUi(self)
        self.splitter.setStretchFactor(0, 0)
        self.splitter.setStretchFactor(1, 100)

        self._colors = colors.colors()
        self.xmin = None
        self.xmax = None
        self.ymin = None
        self.ymax = None
        self._xFunc = None

        # setup class-types, and plotting-curve types for the list
        self.listView_datasets.setItemClass(DFItem)
        self.listView_datasets.supportedClasses = self.supportedClasses

        for xVal in self.x_values:
            self.comboBox_xValues.addItem(xVal[0])
        self.doubleSpinBox_wavelength.setValue(powderdiffraction.CuKa)

        self.plotWidget.plotItem.setLabel("left", "intensity")
        self.plotWidget.plotItem.setLabel("bottom", "two-theta (deg)")

        # --- signal connections --- #
        self.listView_datasets.model().itemsAdded.connect(self.addCurves)
        self.listView_datasets.model().itemTextUpdated.connect(self.plotWidget.plotItem.resetLegend)
        self.listView_datasets.model().itemToggled.connect(self.itemToggled)
        self.listView_datasets.model().rowsMoved.connect(self.processNewLayout)
        self.listView_datasets.model().itemsDeleted.connect(self.processNewLayout)
        # self.listView_datasets.model().rowsRemoved.connect(self.processNewLayout)
        self.doubleSpinBox_wavelength.editingFinished.connect(self.determineXBounds)
        self.doubleSpinBox_wavelength.editingFinished.connect(self.setXDataFunction)
        self.doubleSpinBox_wavelength.editingFinished.connect(self.recalculateXValues)
        self.checkBox_normalize.stateChanged.connect(self.recalculateYValues)
        self.checkBox_xStagger.stateChanged.connect(self.recalculateXValues)
        self.checkBox_xStagger.stateChanged.connect(self.doubleSpinBox_xStagger.setEnabled)
        self.checkBox_yStagger.stateChanged.connect(self.recalculateYValues)
        self.checkBox_yStagger.stateChanged.connect(self.doubleSpinBox_yStagger.setEnabled)
        self.doubleSpinBox_yStagger.valueChanged.connect(self.recalculateYValues)
        self.doubleSpinBox_xStagger.valueChanged.connect(self.recalculateXValues)
        self.comboBox_xValues.currentIndexChanged.connect(self.xSelectionChanged)
        # --- end signal connections --- #

        self.doubleSpinBox_yStagger.setEnabled(False)
        self.doubleSpinBox_xStagger.setEnabled(False)
        self.setXDataFunction()

    def itemToggled(self, item):
        """
        process checking/unchecking of a StructuredDataFrame in the plot

        Parameters
        ----------
        item : DFItem
        """
        if item.checkState:
            item.plotDataItem.setData(
                x=self.xValues(item),
                y=self.yValues(item),
                name=item.text
            )
        else:
            item.plotDataItem.setData(name=None)
            item.plotDataItem.clear()
        self.plotWidget.plotItem.resetLegend()

    def processNewLayout(self):
        self.plotWidget.plotItem.clear()
        self.plotWidget.plotItem.resetLegend()
        self.recalculateYValues()
        for item in self.listView_datasets.iterItems():
            self.plotWidget.addItem(item.plotDataItem)

    def xSelectionChanged(self):
        xIndex = self.comboBox_xValues.currentIndex()
        self.setXDataFunction()
        self.determineXBounds()

        self.doubleSpinBox_wavelength.setEnabled(xIndex not in (1, 2))

        if xIndex == 1:
            xLabel = "Q"
        elif xIndex == 2:
            xLabel = "d-space ({:s})".format(uchars.angstrom)
        else:
            xLabel = "two-theta (deg)"

        self.plotWidget.plotItem.setLabel("bottom", xLabel)

        if xIndex > 2:
            wavelength = self.x_values[xIndex][1]
            self.doubleSpinBox_wavelength.setValue(wavelength)

        self.recalculateXValues()

    def determineXBounds(self):
        # first, where are the xValues coming from
        items = self.listView_datasets.iterItems()
        item1 = next(items)
        xVals = self.xData(item1.df)
        self.xmax = xVals.max()
        self.xmin = xVals.min()
        
        for item in items:
            xVals = self.xData(item.df)
            df_xmax = xVals.max()
            if df_xmax > self.xmax:
                self.xmax = df_xmax
            df_xmin = xVals.min()
            if df_xmin > self.xmin:
                self.xmin = df_xmin

    def determineYBounds(self):
        items = self.listView_datasets.iterItems()
        item1 = next(items)

        self.ymax = item1.df.intensity.max()
        self.ymin = item1.df.intensity.min()

        for item in items:
            df_ymax = item.df.intensity.max()
            if df_ymax > self.ymax:
                self.ymax = df_ymax
            df_ymin = item.df.intensity.min()
            if df_ymin > self.ymin:
                self.ymin = df_ymin

    def determineXYBounds(self):
        self.determineXBounds()
        self.determineYBounds()

    def recalculateXValues(self):
        for item in self.listView_datasets.iterItems():  # type: DFItem
            if item.checkState:
                item.plotDataItem.setData(
                    x=self.xValues(item),
                    y=item.plotDataItem.yData
                )

    def recalculateYValues(self):
        for item in self.listView_datasets.iterItems():  # type: DFItem
            if item.isChecked():
                item.plotDataItem.setData(
                    x=item.plotDataItem.xData,
                    y=self.yValues(item)
                )

    def setXDataFunction(self):
        index = self.comboBox_xValues.currentIndex()
        if index == 1:
            self._xFunc = lambda df: df.Q.values
        elif index == 2:
            self._xFunc = lambda df: df.d_spacing.values
        else:
            wavelength = self.doubleSpinBox_wavelength.value()
            self._xFunc = lambda df: df.twotheta_at_wavelength(wavelength).values

    def xData(self, df):
        """

        Parameters
        ----------
        df : structures.PowderDiffraction

        Returns
        -------
        xVals : np.ndarray
            the data to be used for x-axis plotting

        """
        return self._xFunc(df)

    def yData(self, df):
        """

        Parameters
        ----------
        df : structures.PowderDiffraction

        Returns
        -------
        yVals : np.ndarray
            the data to be used for y-axis plotting
        """
        return df.intensity.values

    def xValues(self, item):
        """

        Parameters
        ----------
        item : DFItem

        Returns
        -------
        xVals : np.ndarray

        """
        xVals = self.xData(item.df)  # get the appropriate xValues, and scale them accordingly

        if self.checkBox_xStagger.isChecked():
            row = self.listView_datasets.row(item)
            stagger_by = row * self.doubleSpinBox_xStagger.value() / 100
            stagger_by *= self.xmax - self.xmin
            xVals += stagger_by

        return xVals

    def yValues(self, item):
        """

        Parameters
        ----------
        item : DFItem

        Returns
        -------
        yVals : np.ndarray

        """
        row = self.listView_datasets.row(item)
        stagger = self.checkBox_yStagger.checkState()
        stagger_by = self.doubleSpinBox_yStagger.value() / 100
        yVals = self.yData(item.df)

        if self.checkBox_normalize.isChecked():
            yMin = yVals.min()
            yShift = yVals - yMin
            yVals = yShift / yShift.max()
            if stagger:
                yVals += stagger_by * row
        else:
            if stagger:
                stagger_by = (self.ymax - self.ymin) * stagger_by
                yVals = yVals + stagger_by * row

        return yVals

    @property
    def nextColor(self):
        return next(self._colors)

    def resetColors(self):
        self._colors = colors.colors()

    def addCurves(self, listItems):
        """
        Parameters
        ----------
        listItems : list of DFItem
        """
        if not listItems:
            return

        self.determineXYBounds()

        for item in listItems:  # type: DFItem
            item.plotDataItem.setData(
                x=self.xValues(item),
                y=self.yValues(item),
                pen=self.nextColor
            )
            self.plotWidget.addItem(item.plotDataItem)


register_visualizations(VisPowderDiffraction)



from . import structures, loaders, plugins
from .structures import StructuredDataFrame
from .loaders import load_file, load_csv

__version__ = '0.1.4'

plugins.import_structures()
plugins.import_loaders()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'C:\Coding\Python\PythonPackageLinks\radie\plugins\visualizations\ui\powderdiffraction.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_PowderDiffraction(object):
    def setupUi(self, PowderDiffraction):
        PowderDiffraction.setObjectName("PowderDiffraction")
        PowderDiffraction.resize(1000, 561)
        self.verticalLayout = QtWidgets.QVBoxLayout(PowderDiffraction)
        self.verticalLayout.setContentsMargins(3, 3, 3, 3)
        self.verticalLayout.setObjectName("verticalLayout")
        self.splitter = QtWidgets.QSplitter(PowderDiffraction)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.layoutWidget = QtWidgets.QWidget(self.splitter)
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayout_left = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout_left.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_left.setObjectName("verticalLayout_left")
        self.listView_datasets = DFListView(self.layoutWidget)
        self.listView_datasets.setObjectName("listView_datasets")
        self.verticalLayout_left.addWidget(self.listView_datasets)
        self.formLayout_parameters = QtWidgets.QFormLayout()
        self.formLayout_parameters.setObjectName("formLayout_parameters")
        self.comboBox_xValues = QtWidgets.QComboBox(self.layoutWidget)
        self.comboBox_xValues.setObjectName("comboBox_xValues")
        self.formLayout_parameters.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.comboBox_xValues)
        self.checkBox_xStagger = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBox_xStagger.setObjectName("checkBox_xStagger")
        self.formLayout_parameters.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.checkBox_xStagger)
        self.checkBox_yStagger = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBox_yStagger.setObjectName("checkBox_yStagger")
        self.formLayout_parameters.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.checkBox_yStagger)
        self.checkBox_normalize = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBox_normalize.setObjectName("checkBox_normalize")
        self.formLayout_parameters.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.checkBox_normalize)
        self.doubleSpinBox_xStagger = QtWidgets.QDoubleSpinBox(self.layoutWidget)
        self.doubleSpinBox_xStagger.setMaximum(1000.0)
        self.doubleSpinBox_xStagger.setSingleStep(5.0)
        self.doubleSpinBox_xStagger.setObjectName("doubleSpinBox_xStagger")
        self.formLayout_parameters.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBox_xStagger)
        self.doubleSpinBox_yStagger = QtWidgets.QDoubleSpinBox(self.layoutWidget)
        self.doubleSpinBox_yStagger.setMaximum(1000.0)
        self.doubleSpinBox_yStagger.setSingleStep(5.0)
        self.doubleSpinBox_yStagger.setObjectName("doubleSpinBox_yStagger")
        self.formLayout_parameters.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBox_yStagger)
        self.doubleSpinBox_wavelength = QtWidgets.QDoubleSpinBox(self.layoutWidget)
        self.doubleSpinBox_wavelength.setDecimals(6)
        self.doubleSpinBox_wavelength.setMinimum(0.01)
        self.doubleSpinBox_wavelength.setSingleStep(0.1)
        self.doubleSpinBox_wavelength.setProperty("value", 1.54)
        self.doubleSpinBox_wavelength.setObjectName("doubleSpinBox_wavelength")
        self.formLayout_parameters.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.doubleSpinBox_wavelength)
        self.verticalLayout_left.addLayout(self.formLayout_parameters)
        self.plotWidget = PlotWidget(self.splitter)
        self.plotWidget.setObjectName("plotWidget")
        self.verticalLayout.addWidget(self.splitter)

        self.retranslateUi(PowderDiffraction)
        QtCore.QMetaObject.connectSlotsByName(PowderDiffraction)

    def retranslateUi(self, PowderDiffraction):
        _translate = QtCore.QCoreApplication.translate
        PowderDiffraction.setWindowTitle(_translate("PowderDiffraction", "PowderDiffraction"))
        self.checkBox_xStagger.setText(_translate("PowderDiffraction", "X-Stagger (%)"))
        self.checkBox_yStagger.setText(_translate("PowderDiffraction", "Y-Stagger (%)"))
        self.checkBox_normalize.setText(_translate("PowderDiffraction", "Normalize intensity"))

from radie.qt.plotlist import DFListView
from radie.qt.plotwidget import PlotWidget



import os
import traceback
import sys
import re

import numpy as np
import pandas as pd

from radie import exceptions
from radie.loaders import Loader, register_loaders
from radie.plugins.structures.vsm import VSM


def load_ideavsm_dat(fname):
    """complete experimental output from Lakeshore, missing sample_id information, so we take it from the file"""
    with open(fname, 'r') as fid:
        file_contents = fid.read()  # type: str

    data_line = re.search("\n\*+ Experiment Data \*+\n", file_contents)
    results_line = re.search("\n\*+ Results \*+\n", file_contents)

    if not data_line:
        raise exceptions.IncorrectFileType("Could not locate the DataLine")
    if not results_line:
        raise exceptions.IncorrectFileType("Could not locate the ResultsLine")

    data_string = file_contents[data_line.end():results_line.start()]  # type: str
    data_lines = data_string.splitlines()[1:-1]
    data_length = int(data_lines[0].split(" ")[1].strip())

    data_segments = (len(data_lines) - 1) / (data_length + 1)  # type: float
    if not data_segments.is_integer():
        # data_segments does not appear to be correctly evaluated as a whole number
        raise exceptions.IncorrectFileType

    df = pd.DataFrame()
    for i in range(int(data_segments)):
        start = 1 + i * (data_length + 1)

        data_label = data_lines[start]  # type: str
        if data_label.endswith(" Data"):
            data_label = data_label[:-5]

        data = np.array(data_lines[start + 1: start + 1 + data_length], dtype="float")
        df[data_label] = data
        name = os.path.basename(fname)
    df.rename(columns={"MomentX": "Moment"}, inplace=True)
    return VSM(df, name=name, date=None)


def load_ideavsm_txt(fname):
    """"simple moment v. Field Output from Lakeshore Idea"""

    try:
        with open(fname, "r") as fid:
            header = [fid.readline() for _ in range(12)]
    except EOFError:
        raise exceptions.IncorrectFileType

    #  check to see if this is an understood filetype
    if not (
        header[0].startswith("Start Time: ") and
        header[1].startswith("Sample ID: ") and
        header[3].startswith("Experiment: ") and
        header[4].startswith("Data File: ") and
        header[9].startswith("***DATA***") and
        header[11].startswith("Field(G)\t Moment(emu)")
    ):
        raise exceptions.IncorrectFileType
    name = header[1].strip().split(":")[1].strip()
    date = header[0].strip().split(":")[1].strip().split()[0]

    df = pd.read_csv(fname, sep="\s+", header=8, index_col=False)
    df.rename(columns={"Field(G)": "Field", "Moment(emu)": "Moment"}, inplace=True)
    df_vsm = VSM(data=df, name=name, date=date)
    return df_vsm


lakeshore_ideas_vsm_dat = Loader(load_ideavsm_dat, VSM, [".dat"], "Lakeshore IDEAs VSM (.dat)")
lakeshore_ideas_vsm_txt = Loader(load_ideavsm_txt, VSM, [".txt"], "Lakeshore IDEAs VSM (.txt)")

register_loaders(
    lakeshore_ideas_vsm_dat,
    lakeshore_ideas_vsm_txt,
)



from collections import OrderedDict

import numpy as np
from radie.structures import StructuredDataFrame, register_data_structures


class PSD(StructuredDataFrame):
    """
    StructuredDataFrame for holding Particle Size Distribution data

    """

    label = "PSD"

    _required_metadata = StructuredDataFrame._required_metadata.copy()
    _required_metadata.update({
        "refractive_index": "1.0",
        "distribution_mode": "volume",
        "ultrasound_time": 0,  # ultrasound time in seconds
    })

    _required_columns = OrderedDict((
        ("diameter", float),
        ("frequency", float),
    ))
    _column_properties = ["oversize"]

    @property
    def d50(self):
        """
        Returns
        -------
        D50 : float
            Particles with diameters of D50 or less account for 50% of the volume
        """
        return self.d(50)

    @property
    def d90(self):
        """
        Returns
        -------
        D90 : float
            Particles with diameters of D90 or less account for 90% of the volume
        """
        return self.d(90)

    @property
    def d10(self):
        """
        Returns
        -------
        d10 : float
            Particles with diameters of D10 or less account for 10% of the volume
        """
        return self.d(10)

    @property
    def dmin(self):
        return self.diameter[self.oversize < 100].iloc[0]

    @property
    def dmax(self):
        return self.diameter[self.oversize > 0.0].iloc[-1]

    def d(self, number):
        """
        Parameters
        ----------
        number : int or float
            Distribution fraction used to determine the diameter.
            Assumes D10 means low diameter end and D90 means high diameter end
            Must be in [0,100] range

        Returns
        -------
        d : float
            Particles with diameters of "d" or less account for "number" of the volume

        """
        if number > 100 or number < 0:
            raise ValueError("Supplied number (%s) must be in [0,100] range" % number)
        if number == 0:
            return self.dmin
        elif number == 100:
            return self.dmax

        oversize_target = 100 - number
        oversize_less = self.oversize <= oversize_target
        oversize_more = self.oversize > oversize_target
        d_min, oversize_min = self.diameter[oversize_more].iloc[-1], self.oversize[oversize_more].iloc[-1]
        d_max, oversize_max = self.diameter[oversize_less].iloc[0], self.oversize[oversize_less].iloc[0]
        return np.exp(
            np.interp(oversize_target, np.array([oversize_max, oversize_min]), np.log(np.array([d_max, d_min]))))

    @property
    def oversize(self):
        """
        Returns
        -------
        oversize : ndarray, float
        """
        oversize = 100.0 - self.frequency.cumsum()
        oversize.name = "oversize"
        return oversize


register_data_structures(PSD)



"""Define loader objects for gsas-type files"""

import io
import os
import sys
import traceback

import numpy as np
import pandas as pd

from radie import exceptions
from radie import loaders
from radie.plugins.structures import powderdiffraction


def parse_gsas_header(fid):
    """a couple of files have the same header structure, so let's read them in one place

    Parameters
    ----------
    fid : io.TextIOBase

    Returns
    -------
    header : dict
    wavelength : float
    data_start_line : int
    bank_line : str

    """

    fid.seek(0)
    headers = dict()
    bank_line = None
    first_data_line = None

    try:
        fid.readline()
    except UnicodeDecodeError:
        raise exceptions.IncorrectFileType

    for i, line in enumerate(fid):
        if line.startswith("#"):
            meta_pair = [value.strip() for value in line[1:].strip().split('=')]
            if len(meta_pair) == 2:
                headers[meta_pair[0]] = meta_pair[1]
        elif line.startswith("BANK"):
            bank_line = line
            first_data_line = i + 2
            break
        else:
            raise exceptions.IncorrectFileType("unrecognized header structure for .fxye files")

    try:
        wavelength = float(headers["Calibrated wavelength"])
    except KeyError:
        raise exceptions.IncorrectFileType("could not find the wavelength")

    return headers, wavelength, first_data_line, bank_line


def load_fxye(fname):
    """.fxye is an Argonne National Labs made file format for powder diffraction that is GSAS compatible

    A more natural csv-type datastructure, as opposed to a GSAS raw file.  Three Columns of data are provided, twotheta
    (in centidegrees), intensity and uncertainty

    Parameters
    ----------
    fname : str
        filename

    Returns
    -------
    powderdiffraction.PowderDiffraction

    Notes
    -----
    from the `11-BM file formats`_ page:

    GSAS supports many formats for powder diffraction data input. The .fxye format for GSAS has an intial header line
    followed by a variable number of comment lines (prefixed by a # character), then an addtional header line. The data
    are listed next in three columns: 1st column is 2θ position (in centidegrees, i.e. degrees × 100), 2nd is intensity,
    and 3rd is standard uncertainty for the intensity values (esd). This format may not be supported by all software
    that claims to read GSAS input files (see GSAS .raw format below)

    When GSAS format files (.fxye or .raw) are requested from the web site, the appropriate GSAS instrument parameter
    file (.prm) is generated and included.

    .. _`11-BM file formats`:
        http://11bm.xray.aps.anl.gov/filetypes.html

    """

    with open(fname, "r") as fid:
        vals = parse_gsas_header(fid)
    headers, wavelength, first_data_line, bank_line = vals

    name = headers.get("User sample name", None)
    if not name:
        name = os.path.basename(fname)

    try:
        df = pd.read_csv(fname, header=None, delimiter=" ", skiprows=first_data_line)
        df.dropna(axis=("index", "columns"), how="all", inplace=True)
    except Exception:
        raise exceptions.IncorrectFileType("\n".join(traceback.format_exception(sys.exc_info())))

    if not len(df.columns) == 3:
        raise exceptions.IncorrectFileType

    if not all(dtype in (float, int) for dtype in df.dtypes):
        raise exceptions.IncorrectFileType

    col_names = dict(zip(df.columns, ("twotheta", "intensity", "uncertainty")))
    df.rename(columns=col_names, inplace=True)
    df["twotheta"] /= 100

    return powderdiffraction.PowderDiffraction(df, name=name, wavelength=wavelength, source="undefined")


def load_raw(fname):
    with open(fname, "r") as fid:
        vals = parse_gsas_header(fid)
        if not vals:
            raise exceptions.IncorrectFileType

        headers, wavelength, first_data_line, bank_line = vals
        bank = bank_line.split()
        try:
            num_points = int(bank[2])
            angle_start = float(bank[5]) / 100
            angle_step = float(bank[6]) / 100
        except Exception:
            raise exceptions.IncorrectFileType

        arr = pd.read_csv(fid, header=None, delim_whitespace=True).values.flatten()[:num_points * 2]

    angle_stop = angle_start + angle_step * (num_points - 1)

    twotheta = np.linspace(angle_start, angle_stop, num_points, dtype=float)
    intensity = arr[0::2]
    uncertainty = arr[1::2]

    name = headers.get("User sample name", None)
    if not name:
        name = os.path.basename(fname)

    return powderdiffraction.PowderDiffraction(
        data=np.vstack((twotheta, intensity, uncertainty)).T,
        columns=["twotheta", "intensity", "uncertainty"],
        name=name,
        wavelength=wavelength
    )


fxye_loader = loaders.Loader(load_fxye, powderdiffraction.PowderDiffraction, (".fxye"), "GSAS .fxye")
gsas_loader = loaders.Loader(load_raw, powderdiffraction.PowderDiffraction,
                             (".raw", ".gsas", ".gsa", ".gs"), "GSAS raw")

loaders.register_loaders(fxye_loader, gsas_loader)



from win32com.client import Dispatch


def createShortcut(path, target='', wDir='', args= '', icon=''):
    shell = Dispatch('WScript.Shell')
    shortcut = shell.CreateShortCut(path)
    shortcut.TargetPath = target
    shortcut.WorkingDirectory = wDir

    if args:
        shortcut.Arguments = args

    if icon == '':
        pass
    else:
        shortcut.IconLocation = icon
    shortcut.save()

path = r"C:\Users\A3R7LZZ\Desktop\Multiline TRL.lnk"
target = r"C:\Miniconda3\pythonw.exe"
args = r"C:\Coding\Python\scikit-rf\qtapps\multiline_trl.py"
createShortcut(path, target, args=args)



from collections import OrderedDict
from radie.structures import StructuredDataFrame, register_data_structures


class DSC(StructuredDataFrame):
    """
    StructuredDataFrame for holding Thermogravimetric analysis data
    Default units are minutes, mg, and celsius
    Please convert results to default units during instantiation, no unit checking is done within
    """

    label = "Differential Scanning Calorimetry"

    _required_metadata = StructuredDataFrame._required_metadata.copy()
    _required_metadata['mass'] = 1. # mg, Default to 1 in case none is provided, used for normalization

    _x = "temperature" # celsius
    _y = "heat_flow" # mW
    _z = "time" # min

    _required_columns = OrderedDict((
        ("temperature", float), # celsius
        ("heat_flow", float), # mW
        ("time", float), # min
    ))
    _column_properties = ["norm_heat_flow"]

    @property
    def norm_heat_flow(self):
        """Normalize the heat flow using the mass found in metadata"""
        return self.heat_flow/self.metadata['mass']

register_data_structures(DSC)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'C:\Coding\Python\PythonPackageLinks\radie\plugins\visualizations\ui\psd.ui'
#
# Created by: PyQt5 UI code generator 5.6
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_PSD(object):
    def setupUi(self, PSD):
        PSD.setObjectName("PSD")
        PSD.resize(1000, 561)
        self.verticalLayout = QtWidgets.QVBoxLayout(PSD)
        self.verticalLayout.setContentsMargins(3, 3, 3, 3)
        self.verticalLayout.setObjectName("verticalLayout")
        self.splitter = QtWidgets.QSplitter(PSD)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.layoutWidget = QtWidgets.QWidget(self.splitter)
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayout_left = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout_left.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_left.setObjectName("verticalLayout_left")
        self.listView_datasets = DFListView(self.layoutWidget)
        self.listView_datasets.setObjectName("listView_datasets")
        self.verticalLayout_left.addWidget(self.listView_datasets)
        self.plotWidget = PlotWidget(self.splitter)
        self.plotWidget.setObjectName("plotWidget")
        self.verticalLayout.addWidget(self.splitter)

        self.retranslateUi(PSD)
        QtCore.QMetaObject.connectSlotsByName(PSD)

    def retranslateUi(self, PSD):
        _translate = QtCore.QCoreApplication.translate
        PSD.setWindowTitle(_translate("PSD", "Particle Size Distribution"))

from radie.qt.plotlist import DFListView
from radie.qt.plotwidget import PlotWidget



from collections import OrderedDict
import uuid
import typing
import json
import os

import pandas
from .. import util


class StructuredDataFrame(pandas.DataFrame):
    """Sub-Class of pandas `DataFrame` that defines data-structures through
       metadata and expected columns of data

    The StructuredDataFrame class is distinguished from a pandas StructuredDataFrame through
    private class variables that provide a way to specify the structure of data,
    which facilitates the easy processing and visualization of that data, e.g.
    in the Radie QtViewer. When creating a subclass of a StructuredDataFrame, the
    following class variables should be defined in the subclass in order to
    define the structure of the subclass.

    _required_columns : OrderedDict()
        a class variable defining the expected data series.  The dict keys are
        the StructuredDataFrame column labels and the values are the expected data types
        (e.g. int, float, etc.)
    _required_metadata : set
        a class variable defining the set of required_metadata keys, used for validation
    _column_properties : OrderedDict()
        a class variable defining the object properties that will return
    _x : typing.Hashable
        the column label for 'x' values, if None defaults to the 1st column
    _y : typing.Hashable
        the column label for 'y' values, if None defaults to the 2nd column
    _z : typing.Hashable
        the column label for 'z' values, if None defaults to the 3rd column

    the pandas `StructuredDataFrame` class has private variable `_metadata`_ which can be
    assigned a list of strings that become instance attributes.  For example, if
    a class class defines :code:`_metadata = ["name", "date"]`, then an instance
    can reference values as :code:`df.name`, and :code:`df.date`.

    Because `_metadata` is more of an ad-hoc feature of `pandas.DataFrame`, and might
    be deprecated in the future, this class uses `_metadata` to define an
    attribute `metadata`, as an `OrderedDict` of metadata that can easily
    interface with the json module for saving.  As such, metadata would be
    accessed by dictionary look-ups, e.g.  :code:`df.metadata["name"]`

    .. _`_metadata`: https://pandas.pydata.org/pandas-docs/stable/internals.html#define-original-properties


    Attributes
    ----------
    metadata : OrderedDict
        an ordered key-value pair for storing all of the metadata, e.g. name,
        date, mass, density, etc.
    x : pandas.Series
        The column of data representing `x` values, default is the first column
    x_col : typing.Hashable
        The column specifier for the `x` values, default is the specifier for
        the first column
    y : pandas.Series
        The column of data representing `y` values, default is the second column
    y_col : typing.Hashable
        The column specifier for the `y` values, default is the specifier for
        the second column
    z : pandas.Series
        The column of data representing `z` values, default is the third column
    z_col : typing.Hashable
        The column specifier for the `z` values, default is the specifier for
        the third column
    uuid : str or None
        a python uuid string for when bookkeeping is necessary, default value is
        None

    """

    label = "StructuredDataFrame"

    _metadata = "metadata"
    _required_metadata = OrderedDict((
        ("name", ""),
        ("date", util.iso_date_string),  # evaluate date-now if none is provided
    ))

    _x = None  # key for default x
    _y = None  # key for default y
    _z = None  # key for default z
    _required_columns = OrderedDict()  # minimum data Series that define the class
    _column_properties = OrderedDict()  # object properties that return a series based on the actual column values
    _loaders = []  # functions that will read in a file and return this class

    def __init__(self, data=None, index=None, columns=None, dtype=None, copy=False, **metadata):
        """The constructor for a StructuredDataFrame, has the same signature as for a Pandas

        Parameters
        ----------
        data
        index
        columns
        dtype
        copy
        metadata
            keyword arguments specifically for the metadata.  Must be JSON compatible

        """
        super(StructuredDataFrame, self).__init__(data, index, columns, dtype, copy)
        self.metadata = OrderedDict()
        self._uuid = None  # type: str

        for meta_key, meta_val in self._required_metadata.items():
            val = metadata.pop(meta_key, meta_val)
            if callable(val):  # can only evaluate functions that take no arguments
                val = val()
            self.metadata[meta_key] = val

        # TODO: implement json checker for metadata
        self.metadata.update(metadata)
        if not self.is_valid():
            raise Exception

    def is_valid(self):
        """determine if the data and metadata match the specifications of the
            StructuredDataFrame class or sublcass

        StructuredDataFrame is meant to be subclassed.  The core specification of a
        StructuredDataFrame are expected columns of data in `._required_columns` and
        also necessary meta-data specified in the `._required_metadata`
        dictionary.  If in the construction or maniuplation of a StructuredDataFrame,
        the object no longer has the required columns or the required metadata,
        it is no longer a valid datastructure.  This function checks the state
        of the instance to see if it matches the class.  All base StructuredDataFrame
        instances should return `True` unless the `name` entry is removed from
        the `.metadata` `dict`.

        Returns
        -------
        bool

        """

        # TODO: implement proper dtype checking
        # dtypes in pandas are not so simple, to start, we're just going to use
        # the consenting adults principle and check only for a column with the
        # correct label, although checking for the correct dtype would be a nice
        # thing to have

        if not self.required_columns().issubset(self.columns):
            return False
        elif not self.required_metadata().issubset(self.metadata):
            return False
        else:
            return True

    @property
    def x(self):
        if not self._x:
            return self.ix[:, 0]
        else:
            return self[self._x]

    @property
    def x_col(self):
        if not self._x:
            return 0
        else:
            return self.columns.get_loc(self._x)

    @property
    def y(self):
        if not self._y:
            return self.ix[:, 1]
        else:
            return self[self._y]

    @property
    def y_col(self):
        if not self._y:
            if len(self.columns) > 1:
                return 1
            else:
                return None
        else:
            return self.columns.get_loc(self._y)

    @property
    def z(self):
        if not self._z:
            return self.ix[:, 2]
        else:
            return self[self._z]

    @property
    def z_col(self):
        if not self._z:
            if len(self.columns) > 2:
                return 2
            else:
                return None
        else:
            return self.columns.get_loc(self._z)

    @classmethod
    def required_metadata(cls):
        return set(cls._required_metadata)

    @classmethod
    def required_columns(cls):
        return set(cls._required_columns)

    @property
    def uuid(self):
        return self._uuid

    def set_uuid(self):
        """
        self.uuid is None by default and must be set.  An error will be
        raised if this function is called twice
        """
        if self.uuid:
            raise AttributeError("self.uuid is already set, specify force_new=True to set a new uuid")
        self._uuid = str(uuid.uuid1())

    def get_uuid(self):
        """get uuid.  If None, set one and then return it"""
        if not self.uuid:
            self.set_uuid()
        return self.uuid

    @classmethod
    def loaders(cls):
        """get a list of all the Loader objects known to return this specific data structure

        Returns
        -------
        list of Loaders

        """
        return cls._loaders

    @classmethod
    def add_loader(cls, loader):
        """append a Loader object to this classes list of known loaders"""
        cls._loaders.append(loader)

    def column_accessors(self):
        """
        return a list of 2-element tuples containing string labels and lambda
        functions (i.e. the accessors) that return column-type values which
        includes the dataframe columns and any properties that return
        column-like values, i.e. the _column_properties.

        Returns
        -------
        accessors : list of tuples

        """

        accessors = list()

        def series_accessor(df, key):
            return lambda: df[key]

        def property_accessor(df, prop):
            return lambda: getattr(df, prop)

        for key in list(self.keys()):
            accessors.append((key, series_accessor(self, key)))

        for prop in self._column_properties:
            accessors.append((prop, property_accessor(self, prop)))

        return accessors

    def savetxt(self, filename, overwrite=False):
        """save an ascii version of the dataframe, with metadata included in comment lines above the datablock

        Parameters
        ----------
        filename : str
        overwrite : bool

        """

        if not filename.endswith(".df"):
            filename += ".df"

        if os.path.isfile(filename) and not overwrite is True:
            raise FileExistsError("You must specify overwrite to be True to overwrite the file")

        with open(filename, "w") as fid:
            fid.write(self.serialize())

    def serialize(self):
        """return a string representation of the StructuredDataFrame, with metadata
            included

        Returns
        -------
        str

        """
        meta = OrderedDict()
        meta["file-type"] = "Radie txt version1"
        meta["class"] = self.__class__.__name__
        meta.update(self.metadata)
        meta_block = "# " + json.dumps(meta, indent=2).replace("\n", "\n# ")
        data_block = self.to_csv(index=None)

        return meta_block + '\n' + data_block

    @classmethod
    def from_clipboard(cls, *args, **kwargs):
        df = pandas.read_clipboard()
        # TODO: figure out a clever way to parse required columns, ignore for now for the sake of convenience
        return cls(data=df, *args, **kwargs)

    @property
    def _constructor(self):
        return self.__class__

    def rdplot(self, *args, **kwargs):
        """Subclasses may implement custom plotting functionality"""
        raise NotImplementedError("only available for specific subclasses of StructuredDataFrame")




"""Module for the Visualizations of Thermogravimetric Analysis Datasets"""

import os

import pyqtgraph as pg
import numpy as np

from radie.qt import cfg, colors, uchars
from radie.qt import plotlist
from radie.qt.visualizations import base, register_visualizations
from radie.plugins.structures import tga
from radie.plugins.visualizations.ui.tga import Ui_TGA


class DFItem(plotlist.DFItem):
    """an item class with pyqtgraph xy curve handles of type PlotDataItem"""

    def __init__(self, ref, item_list, name=None):
        super(DFItem, self).__init__(ref, item_list, name)
        self.plotDataItemY1 = pg.PlotDataItem()
        self.plotDataItemY1.setData(name=self.text)
        self.plotDataItemY2 = pg.PlotDataItem()
        self.plotDataItemY2.setData(name=self.text)

    def setText(self, value):
        self.text = value
        self.plotDataItemY1.setData(name=value)
        self.plotDataItemY2.setData(name=value)


class VisTGA(base.Visualization, Ui_TGA):
    name = "TGA"
    description = "A Visualization to compare thermogravimetric analysis results"
    _icon_image = os.path.join(os.path.dirname(__file__), "icons/tga_chart_curve.png")

    supportedClasses = (
        tga.TGA,
    )

    X_COMBO = ['Temperature', 'Time']
    Y1_COMBO = ['Weight', 'Deriv. Weight', 'Temperature']
    Y2_COMBO = ['None', 'Deriv. Weight', 'Temperature']

    def __init__(self, parent=None):
        super(VisTGA, self).__init__(parent)
        self.setupUi(self)
        self.splitter.setStretchFactor(0, 0)
        self.splitter.setStretchFactor(1, 100)

        self._colors = colors.colors()
        self.xmin = None
        self.xmax = None
        self.y1min = None
        self.y1max = None
        self.y2min = None
        self.y2max = None

        self.xlabel = None
        self.y1label = None
        self.y2label = None

        self._xFunc = None
        self._y1Func = None
        self._y2Func = None

        # setup class-types, and plotting-curve types for the list
        self.listView_datasets.supportedClasses = self.supportedClasses
        self.listView_datasets.setItemClass(DFItem)

        # Setup plotting choices
        for x_key in self.X_COMBO:
            self.comboBox_x.addItem(x_key)
        for y_key in self.Y1_COMBO:
            self.comboBox_y1.addItem(y_key)
        for y_key in self.Y2_COMBO:
            self.comboBox_y2.addItem(y_key)

        self.comboBox_x.setCurrentIndex(0)
        self.comboBox_y1.setCurrentIndex(0)
        self.comboBox_y2.setCurrentIndex(1)

        self.plotWidget.plotItem.setLabel("left", "Weight (mg)")
        self.plotWidget.plotItem.setLabel("bottom", "Temperature (°C)")

        # Setup secondary y axis
        # pyqtgraph does not have this really built in
        # A new viewbox must be created which will hold its own plotDataItems
        # That viewbox must be linked to the existing plotItem
        # Adapted from : https://stackoverflow.com/questions/23679159/two-y-scales-in-pyqtgraph-twinx-like
        # PlotDataItems (or PlotCurveItems etc..) can now be added to
        # the self.y2ViewBox via its addItem function
        self.plotWidget.plotItem.showAxis("right")
        self.plotWidget.plotItem.setLabel("right", "Deriv. Weight (mg/°C)")
        self.y2ViewBox = pg.ViewBox()
        self.plotWidget.plotItem.scene().addItem(self.y2ViewBox)
        self.plotWidget.plotItem.getAxis('right').linkToView(self.y2ViewBox)
        self.y2ViewBox.setXLink(self.plotWidget.plotItem)
        self.updateViews()
        self.plotWidget.plotItem.getViewBox().sigResized.connect(self.updateViews)

        # --- signal connections --- #
        # Taking the lazy approach with a complete redraw every time something changes
        # as opposed to changing only what is required
        self.listView_datasets.model().itemsAdded.connect(self.processNewLayout)
        self.listView_datasets.model().itemTextUpdated.connect(self.processNewLayout)
        self.listView_datasets.model().itemToggled.connect(self.processNewLayout)
        self.listView_datasets.model().rowsMoved.connect(self.processNewLayout)
        self.listView_datasets.model().itemsDeleted.connect(self.processNewLayout)
        self.checkBox_normalizeWeight.stateChanged.connect(self.processNewLayout)
        self.comboBox_x.currentIndexChanged.connect(self.processNewLayout)
        self.comboBox_y1.currentIndexChanged.connect(self.processNewLayout)
        self.comboBox_y2.currentIndexChanged.connect(self.processNewLayout)
        # --- end signal connections --- #

        self.setXDataFunction()
        self.setY1DataFunction()
        self.setY2DataFunction()

    def updateViews(self):
        """Needed for the y2 axis"""
        self.y2ViewBox.setGeometry(self.plotWidget.plotItem.getViewBox().sceneBoundingRect())
        self.y2ViewBox.linkedViewChanged(self.plotWidget.plotItem.getViewBox(), self.y2ViewBox.XAxis)

    def processNewLayout(self):
        self.plotWidget.plotItem.clear()
        self.plotWidget.plotItem.resetLegend()
        self.y2ViewBox.clear()

        self.setXDataFunction()
        self.setY1DataFunction()
        self.setY2DataFunction()

        self.setAxisLabels()

        self.addCurves([item for item in self.listView_datasets.iterItems() if item.checkState])

    def setAxisLabels(self):
        """
        Set the left, bottom and right axis labels based on the selection in the combo boxes
        Returns
        -------

        """
        # get and set the bottom axis label
        index = self.comboBox_x.currentIndex()
        if index == 0:
            self.xlabel = "Temperature (°C)"
        elif index == 1:
            self.xlabel = "Time (min)"
        self.plotWidget.plotItem.setLabel("bottom", self.xlabel)

        # get and set the left axis label
        index = self.comboBox_y1.currentIndex()
        norm = self.checkBox_normalizeWeight.isChecked()
        if index == 0:
            self.y1label = 'Weight ({})'.format('%' if norm else 'mg')
        elif index == 1:
            self.y1label = 'Deriv. Weight ({}/°C)'.format('%' if norm else 'mg')
        elif index == 2:
            self.y1label = 'Temperature (°C)'
        self.plotWidget.plotItem.setLabel("left", self.y1label)

        # get and set the right axis label
        index = self.comboBox_y2.currentIndex()
        norm = self.checkBox_normalizeWeight.isChecked()
        if index == 0:
            # is this really he best way to do control values on axis??
            self.plotWidget.plotItem.getAxis('right').style['showValues'] = False
            self.plotWidget.plotItem.getAxis('right').setGrid(False)
            self.y2label = ''
        elif index == 1:
            # is this really he best way to do control values on axis??
            self.plotWidget.plotItem.getAxis('right').style['showValues'] = True
            self.plotWidget.plotItem.getAxis('right').setGrid(True)
            self.y2label = 'Deriv. Weight ({}/°C)'.format('%' if norm else 'mg')
        elif index == 2:
            # is this really he best way to do control values on axis??
            self.plotWidget.plotItem.getAxis('right').style['showValues'] = True
            self.plotWidget.plotItem.getAxis('right').setGrid(True)
            self.y2label = 'Temperature (°C)'
        self.plotWidget.plotItem.showAxis("right")
        self.plotWidget.plotItem.setLabel("right", self.y2label)

    def hasY2(self):
        return self.comboBox_y2.currentIndex() > 0

    def setXDataFunction(self):
        index = self.comboBox_x.currentIndex()
        if index == 0:
            self._xFunc = lambda df: df.temperature.values
        elif index == 1:
            self._xFunc = lambda df: df.time.values

    def setY1DataFunction(self):
        index = self.comboBox_y1.currentIndex()
        norm = self.checkBox_normalizeWeight.isChecked()
        if index == 0:
            if norm:
                self._y1Func = lambda df: df.norm_weight.values
            else:
                self._y1Func = lambda df: df.weight.values
        elif index == 1:
            if norm:
                self._y1Func = lambda df: df.deriv_norm_weight.values
            else:
                self._y1Func = lambda df: df.deriv_weight.values
        elif index == 2:
            self._y1Func = lambda df: df.temperature.values

    def setY2DataFunction(self):
        index = self.comboBox_y2.currentIndex()
        norm = self.checkBox_normalizeWeight.isChecked()
        if index == 0:
            self._y2Func = lambda df: np.array([])
        elif index == 1:
            if norm:
                self._y2Func = lambda df: df.deriv_norm_weight.values
            else:
                self._y2Func = lambda df: df.deriv_weight.values
        elif index == 2:
            self._y2Func = lambda df: df.temperature.values

    def xData(self, df):
        """

        Parameters
        ----------
        df : structures.PowderDiffraction

        Returns
        -------
        xVals : np.ndarray
            the data to be used for x-axis plotting

        """
        return self._xFunc(df)

    def y1Data(self, df):
        """

        Parameters
        ----------
        df : structures.PowderDiffraction

        Returns
        -------
        yVals : np.ndarray
            the data to be used for y-axis plotting
        """
        return self._y1Func(df)

    def y2Data(self, df):
        return self._y2Func(df)

    def xValues(self, item):
        """

        Parameters
        ----------
        item : DFItem

        Returns
        -------
        xVals : np.ndarray

        """
        xVals = self.xData(item.df)  # get the appropriate xValues, and scale them accordingly
        return xVals

    def y1Values(self, item):
        """

        Parameters
        ----------
        item : DFItem

        Returns
        -------
        yVals : np.ndarray

        """
        y1Vals = self.y1Data(item.df)
        return y1Vals

    def y2Values(self, item):
        """

        Parameters
        ----------
        item : DFItem

        Returns
        -------
        yVals : np.ndarray

        """
        y2Vals = self.y2Data(item.df)
        return y2Vals

    @property
    def nextColor(self):
        return next(self._colors)

    def resetColors(self):
        self._colors = colors.colors()

    def addCurves(self, listItems):
        """
        Parameters
        ----------
        listItems : list of DFItem
        """
        if not listItems:
            return

        self.resetColors()

        for item in listItems:  # type: DFItem

            # Color attribution before checkState allows constant colors during toggling
            color1 = self.nextColor
            color2 = self.nextColor

            if item.checkState:
                item.plotDataItemY1.setData(
                    x=self.xValues(item),
                    y=self.y1Values(item),
                    pen=color1
                )
                self.plotWidget.addItem(item.plotDataItemY1)

                if self.hasY2():
                    item.plotDataItemY2.setData(
                        x=self.xValues(item),
                        y=self.y2Values(item),
                        pen=color2
                    )
                    self.y2ViewBox.addItem(item.plotDataItemY2)
                    # Need to manually add the item to the legend since it is in a different viewbox
                    self.plotWidget.plotItem.legend.addItem(item.plotDataItemY2,
                                                            '{} y2'.format(item.plotDataItemY2.name()))



register_visualizations(VisTGA)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'C:\Users\a2nldzz\Documents\repos\radie\radie\plugins\visualizations\ui\tga.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_TGA(object):
    def setupUi(self, TGA):
        TGA.setObjectName("TGA")
        TGA.resize(1000, 561)
        self.verticalLayout = QtWidgets.QVBoxLayout(TGA)
        self.verticalLayout.setContentsMargins(3, 3, 3, 3)
        self.verticalLayout.setObjectName("verticalLayout")
        self.splitter = QtWidgets.QSplitter(TGA)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.layoutWidget = QtWidgets.QWidget(self.splitter)
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayout_left = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout_left.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_left.setObjectName("verticalLayout_left")
        self.listView_datasets = DFListView(self.layoutWidget)
        self.listView_datasets.setObjectName("listView_datasets")
        self.verticalLayout_left.addWidget(self.listView_datasets)
        self.formLayout_parameters = QtWidgets.QFormLayout()
        self.formLayout_parameters.setObjectName("formLayout_parameters")
        self.comboBox_y1 = QtWidgets.QComboBox(self.layoutWidget)
        self.comboBox_y1.setObjectName("comboBox_y1")
        self.formLayout_parameters.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.comboBox_y1)
        self.label_2 = QtWidgets.QLabel(self.layoutWidget)
        self.label_2.setObjectName("label_2")
        self.formLayout_parameters.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.label_2)
        self.comboBox_y2 = QtWidgets.QComboBox(self.layoutWidget)
        self.comboBox_y2.setObjectName("comboBox_y2")
        self.formLayout_parameters.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.comboBox_y2)
        self.label_3 = QtWidgets.QLabel(self.layoutWidget)
        self.label_3.setObjectName("label_3")
        self.formLayout_parameters.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.label_3)
        self.checkBox_normalizeWeight = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBox_normalizeWeight.setObjectName("checkBox_normalizeWeight")
        self.formLayout_parameters.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.checkBox_normalizeWeight)
        self.comboBox_x = QtWidgets.QComboBox(self.layoutWidget)
        self.comboBox_x.setObjectName("comboBox_x")
        self.formLayout_parameters.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.comboBox_x)
        self.label = QtWidgets.QLabel(self.layoutWidget)
        self.label.setObjectName("label")
        self.formLayout_parameters.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.label)
        self.verticalLayout_left.addLayout(self.formLayout_parameters)
        self.plotWidget = PlotWidget(self.splitter)
        self.plotWidget.setObjectName("plotWidget")
        self.verticalLayout.addWidget(self.splitter)

        self.retranslateUi(TGA)
        QtCore.QMetaObject.connectSlotsByName(TGA)

    def retranslateUi(self, TGA):
        _translate = QtCore.QCoreApplication.translate
        TGA.setWindowTitle(_translate("TGA", "TGA"))
        self.label_2.setText(_translate("TGA", "Primary Vertical Axis"))
        self.label_3.setText(_translate("TGA", "Secondary Vertical Axis"))
        self.checkBox_normalizeWeight.setText(_translate("TGA", "Normalize weight"))
        self.label.setText(_translate("TGA", "Horizontal Axis"))

from radie.qt.plotlist import DFListView
from radie.qt.plotwidget import PlotWidget



"""module for the visualizations of generic datastructures"""
import os

from PyQt5 import QtGui, QtWidgets, QtCore
import pyqtgraph as pg

from . import base
from .. import dpi
from .. import cfg, colors
from .. import plotlist
from .. import functions as fn
from ..plotlist import DFXYListView
from ..plotwidget import PlotWidget


class DFItem(plotlist.DFItem):
    """an item class with pyqtgraph xy curve handles of type PlotDataItem"""
    def __init__(self, ref, item_list, name=None):
        super(DFItem, self).__init__(ref, item_list, name)
        self.plotDataItem = pg.PlotDataItem()
        self.color = None

        self.legend_name = True
        self.legend_label = False
        self.legend_ycol = True

        self.plotDataItem.setData(
            x=self.x_data(),
            y=self.y_data(),
            name=self.text
        )

    def setLegend(self, legend_name=None, legend_label=None, legend_ycol=None):
        legend = list()

        if type(legend_name) is bool:
            self.legend_name = legend_name
            if legend_name:
                legend.append(self.ref.df.metadata["name"])
        if type(legend_label) is bool:
            self.legend_label = legend_label
            if legend_label:
                legend.append(self.text)
        if type(legend_ycol) is bool:
            self.legend_ycol = legend_ycol
            if legend_ycol:
                legend.append(str(self.y_accessor))

        self.plotDataItem.setData(name=" - ".join(legend))

    def setText(self, value):
        self.text = value
        if self.legend_label:  # only update if the legend is showing this data
            self.setLegend()


class XYScatter(base.Visualization):
    """A generic XY scatter visualization

    This visualization is meant to accomodate all possible data-frames.  It provides a combo-box in the list widget
    to select which columns of data will be used for the X and Y values, which will be non-sensical in some cases

    """
    name = "XY Scatter"
    description = "A generic visualization of XY curves from StructuredDataFrame Series"
    _icon_image = os.path.join(cfg.icon_path, "xyscatter.svg")

    def setupUi(self):
        self.resize(1000, 560)
        self.setWindowTitle('DataFrames')
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self)
        self.verticalLayout_2.setContentsMargins(3, 3, 3, 3)
        self.splitter = QtWidgets.QSplitter(self)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.layoutWidget = QtWidgets.QWidget(self.splitter)
        self.verticalLayout = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.treeView_datasets = DFXYListView(self.layoutWidget)
        self.verticalLayout.addWidget(self.treeView_datasets)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.comboBox_lineStyle = QtWidgets.QComboBox(self.layoutWidget)
        self.comboBox_lineStyle.addItem("line")
        self.comboBox_lineStyle.addItem("points")
        self.horizontalLayout.addWidget(self.comboBox_lineStyle)
        self.label_alpha = QtWidgets.QLabel('Scatter Alpha', self.layoutWidget)
        self.horizontalLayout.addWidget(self.label_alpha)
        self.slider_alpha = QtWidgets.QSlider(self.layoutWidget)
        self.slider_alpha.setMaximum(255)
        self.slider_alpha.setProperty("value", 255)
        self.slider_alpha.setOrientation(QtCore.Qt.Horizontal)
        self.horizontalLayout.addWidget(self.slider_alpha)
        self.label_alphaValue = QtWidgets.QLabel('255', self.layoutWidget)
        self.horizontalLayout.addWidget(self.label_alphaValue)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.groupBox_Legend = QtWidgets.QGroupBox('Legend', self.layoutWidget)
        self.groupBox_Legend.setCheckable(True)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.groupBox_Legend)
        self.checkBox_legend_dfname = QtWidgets.QCheckBox('DF Name', self.groupBox_Legend)
        self.checkBox_legend_dfname.setChecked(True)
        self.horizontalLayout_2.addWidget(self.checkBox_legend_dfname)
        self.checkBox_legend_label = QtWidgets.QCheckBox('Item Label', self.groupBox_Legend)
        self.checkBox_legend_label.setChecked(False)
        self.horizontalLayout_2.addWidget(self.checkBox_legend_label)
        self.checkBox_legend_ycolumn = QtWidgets.QCheckBox('Y-Column', self.groupBox_Legend)
        self.checkBox_legend_ycolumn.setChecked(True)
        self.horizontalLayout_2.addWidget(self.checkBox_legend_ycolumn)
        self.verticalLayout.addWidget(self.groupBox_Legend)
        self.formLayout_plotOptions = QtWidgets.QFormLayout()
        self.label_xlabel = QtWidgets.QLabel('X-Label', self.layoutWidget)
        self.formLayout_plotOptions.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.label_xlabel)
        self.lineEdit_xlabel = QtWidgets.QLineEdit(self.layoutWidget)
        self.formLayout_plotOptions.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.lineEdit_xlabel)
        self.label_ylabel = QtWidgets.QLabel('Y-Label', self.layoutWidget)
        self.formLayout_plotOptions.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.label_ylabel)
        self.lineEdit_ylabel = QtWidgets.QLineEdit(self.layoutWidget)
        self.formLayout_plotOptions.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.lineEdit_ylabel)
        self.verticalLayout.addLayout(self.formLayout_plotOptions)
        self.plotWidget = PlotWidget(self.splitter)
        self.verticalLayout_2.addWidget(self.splitter)

    def __init__(self, parent=None):
        super(XYScatter, self).__init__(parent)
        self.setupUi()
        self.splitter.setStretchFactor(0, 0)
        self.splitter.setStretchFactor(1, 1)

        self.splitter.setSizes(dpi.width_by_height(280, 720))

        self.comboBox_lineStyle.setItemIcon(0, QtGui.QIcon(os.path.join(cfg.icon_path, "xyline.svg")))
        self.comboBox_lineStyle.setItemIcon(1, QtGui.QIcon(os.path.join(cfg.icon_path, "xypoints.svg")))

        self.treeView_datasets.setItemClass(DFItem)
        self.treeView_datasets.supportedClasses = self.supportedClasses

        self.treeView_datasets.model().itemsAdded.connect(self.addCurves)
        self.treeView_datasets.model().itemAccessorChanged.connect(self.itemDataChanged)
        self.treeView_datasets.model().itemAccessorChanged.connect(self.legendSelectionChanged)
        self.treeView_datasets.model().itemsDeleted.connect(self.processNewLayout)
        self.treeView_datasets.model().rowsMoved.connect(self.processNewLayout)
        self.treeView_datasets.model().itemToggled.connect(self.itemToggled)
        self.treeView_datasets.model().itemTextUpdated.connect(self.legendSelectionChanged)
        self.comboBox_lineStyle.currentIndexChanged.connect(self.setLineStyle)
        self.lineEdit_xlabel.textChanged.connect(self.setXLabel)
        self.lineEdit_ylabel.textChanged.connect(self.setYLabel)

        for checkBox in (self.checkBox_legend_dfname, self.checkBox_legend_label, self.checkBox_legend_ycolumn):
            checkBox.stateChanged.connect(self.legendSelectionChanged)
        self.groupBox_Legend.toggled.connect(self.toggleLegend)

        self._colors = None
        self.resetColors()

        self.alphaChanged = pg.SignalProxy(self.slider_alpha.valueChanged, slot=self.setAlpha, delay=0.05)
        self.slider_alpha.valueChanged.connect(self.label_alphaValue.setNum)

    def legendSelectionChanged(self):
        for item in self.treeView_datasets.model().dflist:  # type: DFItem
            item.setLegend(
                self.checkBox_legend_dfname.isChecked(),
                self.checkBox_legend_label.isChecked(),
                self.checkBox_legend_ycolumn.isChecked(),
            )
            self.plotWidget.plotItem.resetLegend()

    def toggleLegend(self, on: bool):
        if on:
            self.plotWidget.plotItem.resetLegend()
        else:
            self.plotWidget.plotItem.removeLegend()

    def setAlpha(self):
        if not self.useSymbols():
            return
        for item in self.treeView_datasets.iterItems():
            color = "{:s}{:02x}".format(item.color, self.slider_alpha.value())
            item.plotDataItem.setSymbolPen(color)

    def useSymbols(self):
        return bool(self.comboBox_lineStyle.currentIndex())

    def setAxisLabel(self, axis: str, text: str):
        if not text:
            self.plotWidget.plotItem.showLabel(axis, False)
            return
        self.plotWidget.setLabel(axis=axis, text=text)

    def setXLabel(self, text: str):
        self.setAxisLabel("bottom", text)

    def setYLabel(self, text: str):
        self.setAxisLabel("left", text)

    def setLineStyle(self, checkState: bool):
        if self.useSymbols():
            for item in self.treeView_datasets.iterItems():  # type: DFItem
                color = "{:s}{:02x}".format(item.color, self.slider_alpha.value())
                item.plotDataItem.setData(symbolPen=color, symbol='o', symbolSize=5)
                item.plotDataItem.setPen(None)
        else:
            for item in self.treeView_datasets.iterItems():  # type: DFItem
                item.plotDataItem.setData(symbolPen=None, symbol=None)
                item.plotDataItem.setPen(item.color)

    def nextColor(self):
        return next(self._colors)

    def resetColors(self):
        self._colors = colors.colors()

    def itemDataChanged(self, item: DFItem):
        if not item.isChecked():
            return

        item.plotDataItem.setData(
            x=item.x_data(),
            y=item.y_data()
        )

    def processNewLayout(self):
        self.plotWidget.plotItem.clear()
        self.plotWidget.plotItem.resetLegend()
        for item in self.treeView_datasets.iterItems():
            self.plotWidget.addItem(item.plotDataItem)

    def itemToggled(self, item):
        """process checking/unchecking of a StructuredDataFrame in the plot

        Parameters
        ----------
        item : DFItem

        """
        if item.checkState:
            item.plotDataItem.setData(
                x=item.x_data(),
                y=item.y_data(),
            )
            item.setLegend(self.checkBox_legend_dfname.isChecked(), self.checkBox_legend_label.isChecked(),
                           self.checkBox_legend_ycolumn.isChecked())
        else:
            item.plotDataItem.setData(name=None)
            item.plotDataItem.clear()
        self.plotWidget.plotItem.resetLegend()

    def addCurves(self, items):
        """
        Parameters
        ----------
        items : list of DFItem
        """
        if self.useSymbols():
            for item in items:  # type: DFItem
                item.color = self.nextColor()
                item.plotDataItem.setData(pen=None, symbol='o', symbolPen=item.color, symbolBrush=None)
                item.setLegend(self.checkBox_legend_dfname.isChecked(), self.checkBox_legend_label.isChecked(),
                               self.checkBox_legend_ycolumn.isChecked())
                item.plotDataItem.updateItems()
                self.plotWidget.plotItem.addItem(item.plotDataItem)
        else:
            alpha = self.slider_alpha.value()
            for item in items:
                item.color = self.nextColor()
                color = "{:s}{:02x}".format(item.color, alpha)
                item.plotDataItem.setData(pen=color, symbol=None, symbolPen=None, symbolBrush=None)
                item.setLegend(self.checkBox_legend_dfname.isChecked(), self.checkBox_legend_label.isChecked(),
                               self.checkBox_legend_ycolumn.isChecked())
                item.plotDataItem.updateItems()
                self.plotWidget.plotItem.addItem(item.plotDataItem)


def test():
    import sys
    from ...plugins import examples
    from ..masterdftree import DFReference
    app = fn.instantiate_app()
    fn.reset_excepthook()
    cfg.set_dpi_scaling()

    df = examples.example_powderdiffraction()
    ref1 = DFReference(df, None)

    plot = XYScatter()
    plot.treeView_datasets.addDataFrames(ref1, ref1, ref1)
    plot.show()

    sys.exit(app.exec_())



"""module to define dflistview, using the QAbstractListModel"""
import typing

import numpy as np
from PyQt5 import QtCore, QtGui, QtWidgets

from ..structures import StructuredDataFrame
from .indexabledict import IndexableDict
from .errors import DFTypeError
from .masterdftree import DFReference
from . import functions as fn


class DFItem(object):
    def __init__(self, ref, item_list, name=None):
        """
        Parameters
        ----------
        ref : DFReference
        item_list : DFItemList
        name : str
        """
        self.ref = ref
        self.list = item_list
        self.checkState = QtCore.Qt.Checked
        self.text = name if name else self.ref.df.metadata["name"]
        self.accessors = IndexableDict(self.ref.df.column_accessors())
        self.x_accessor = self.accessors.getKey(self.ref.df.x_col)
        y_idx = self.ref.df.y_col
        if y_idx is None:
            self.y_accessor = None
        else:
            self.y_accessor = self.accessors.getKey(y_idx)
        z_idx = self.ref.df.z_col
        if z_idx is None:
            self.z_accessor = None
        else:
            self.z_accessor = self.accessors.getKey(z_idx)

    @property
    def df(self):
        return self.ref.df

    def x_data(self) -> np.ndarray:
        data = self.accessors[self.x_accessor]().values
        return data

    def y_data(self) -> np.ndarray:
        return self.accessors[self.y_accessor]().values

    def z_data(self) -> np.ndarray:
        return self.accessors[self.z_accessor]().values

    def setText(self, value):
        self.text = value

    def setCheckState(self, value):
        self.checkState = value

    def isChecked(self):
        return bool(self.checkState)


class DFItemList(list):
    def __init__(self, iterable=None):
        super(DFItemList, self).__init__()
        if not iterable:
            return

        for item in iterable:
            if not isinstance(object, DFItem):
                raise TypeError("Only DFReference objects are allowed")
            self.append(item)

        self._itemClass = None

    @property
    def itemClass(self):
        if not self._itemClass:
            raise AttributeError("itemClass attribute not yet assigned")
        return self._itemClass

    @itemClass.setter
    def itemClass(self, cls):
        if not issubclass(cls, DFItem):
            raise TypeError("cls must be a DFItem subclass not {:}".format(cls.__name__))
        self._itemClass = cls

    def append(self, ref):
        """append a new StructuredDataFrame to the List

        Parameters
        ----------
        ref : DFReference
        """
        if not isinstance(ref, DFReference):
            raise TypeError("Only DFReference objects are allowed, not type: {:}".format(type(DFReference)))
        item = self.itemClass(ref, self)
        super(DFItemList, self).append(item)

    def __setitem__(self, key, value):
        if not isinstance(value, DFReference):
            raise TypeError("Only DFReference objects are allowed")
        super(DFItemList, self).__setitem__(key, value)


class DFListModel(QtCore.QAbstractItemModel):
    """
    Define a generic Model for use with lists of DataFrames to be plotted.  This model is a QAbstractItemModel
    meant to define the behavior for a QListView.  It can however also be used with a QTreeView, but will probably
    require some custom behavior via a sub-class
    """

    itemsAdded = QtCore.pyqtSignal(object)
    itemToggled = QtCore.pyqtSignal(object)
    itemTextUpdated = QtCore.pyqtSignal(object)
    itemAccessorChanged = QtCore.pyqtSignal(object)
    itemsDeleted = QtCore.pyqtSignal()

    def __init__(self, dflist: DFItemList):
        super(DFListModel, self).__init__()
        if not isinstance(dflist, DFItemList):
            raise TypeError("DFListModel only supports DFItemLists")
        self.dflist = dflist  # type: DFItemList
        self._headers = ["Name", "x-data", "y-data"]
        self._columnCount = len(self._headers)

    def setColumnCount(self, value):
        self._columnCount = int(value)

    def columnCount(self, parent: QtCore.QModelIndex = None):
        return self._columnCount

    def setHeaders(self, headers: list):
        headers = [str(header) for header in headers]
        self._headers = headers
        self._columnCount = len(self._headers)

    def headerData(self, section: int, orientation: QtCore.Qt.Orientation, role: int = ...):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return self._headers[section]
        else:
            return None

    def index(self, row: int, column: int, parent: QtCore.QModelIndex = None) -> QtCore.QModelIndex:
        if not parent:
            parent = QtCore.QModelIndex()

        if not self.hasIndex(row, column, parent):
            return QtCore.QModelIndex()

        index = self.createIndex(row, column, self.dflist)
        return index

    def parent(self, child: QtCore.QModelIndex):
        return QtCore.QModelIndex()

    def hasChildren(self, parent: QtCore.QModelIndex = None):
        if parent.row() == -1:
            return True
        else:
            return False

    def rowCount(self, parent: QtCore.QModelIndex = ...):
        return len(self.dflist)

    def data(self, index: QtCore.QModelIndex, role: int = ...):
        col = index.column()
        if col == 0:
            if role == QtCore.Qt.CheckStateRole:
                item = self.dflist[index.row()]  # type: DFItem
                return item.checkState
            elif role in (QtCore.Qt.EditRole, QtCore.Qt.DisplayRole):
                item = self.dflist[index.row()]  # type: DFItem
                return item.text

        elif col > 0 and role in (QtCore.Qt.EditRole, QtCore.Qt.DisplayRole):
            if col == 1:
                return str(self.dflist[index.row()].x_accessor)
            if col == 2:
                return str(self.dflist[index.row()].y_accessor)
            if col == 3:
                return str(self.dflist[index.row()].z_accessor)

        return QtCore.QVariant()

    def setData(self, index: QtCore.QModelIndex, value: typing.Any, role: int):
        item = self.dflist[index.row()]  # type: DFItem
        col = index.column()
        if role == QtCore.Qt.CheckStateRole:
            item.setCheckState(value)
            self.itemToggled.emit(item)
            return True
        elif role == QtCore.Qt.EditRole:
            if col == 0:
                item.setText(str(value))
                self.itemTextUpdated.emit(item)
                return True

            else:
                if col == 1:
                    item.x_accessor = value
                elif col == 2:
                    item.y_accessor = value
                elif col == 3:
                    item.z_accessor = value
                else:
                    return False
                self.itemAccessorChanged.emit(item)
                return True
        return False

    def flags(self, index: QtCore.QModelIndex):
        col = index.column()
        if col == 0:
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEditable | \
                   QtCore.Qt.ItemIsDropEnabled | QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsUserCheckable
        else:
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEditable | \
                   QtCore.Qt.ItemIsDropEnabled

    def removeRow(self, row: int, parent: QtCore.QModelIndex=None):
        if not parent:
            parent = QtCore.QModelIndex()
        self.beginRemoveRows(parent, row, row)
        del self.dflist[row]
        self.endRemoveRows()

    def removeRows(self, row: int, count: int, parent: QtCore.QModelIndex=None):
        if not parent:
            parent = QtCore.QModelIndex()
        first = row
        last = row + count - 1
        self.beginRemoveRows(parent, first, last)  # last row inclusive
        del self.dflist[row:last + 1]  # last row not inclusive
        self.endRemoveRows()

    def internalMove(self, dropIndex: QtCore.QModelIndex, movingIndexes: list):
        indexes = sorted(movingIndexes, key=lambda index: index.row())
        parent = QtCore.QModelIndex()

        first_row = indexes[0].row()
        last_row = indexes[-1].row()

        insert_row = dropIndex.row()
        if insert_row < 0:  # dragging to the end yields `insert_row = -1`
            insert_row = len(self.dflist)

        if first_row <= insert_row <= last_row + 1:
            return  # drop spot is in-between somewhere, no purpose in moving
            # and in fact and exception will occur if you try to move
            # and there is no change in the list

        if first_row > insert_row:
            self.beginMoveRows(parent, indexes[0].row(), insert_row, parent, insert_row)
            for index in indexes:  # type: QtCore.QModelIndex
                row = index.row()
                self.dflist.insert(insert_row, self.dflist.pop(row))
                insert_row += 1
        else:
            self.beginMoveRows(parent, insert_row, indexes[-1].row(), parent, insert_row)
            for index in indexes[::-1]:  # type: QtCore.QModelIndex
                row = index.row()
                self.dflist.insert(insert_row - 1, self.dflist.pop(row))
                insert_row -= 1
        self.endMoveRows()

    def addDataFrames(self, *refs):
        first = len(self.dflist)
        last = len(refs) - 1
        self.beginInsertRows(QtCore.QModelIndex(), first, last)
        for ref in refs:
            self.dflist.append(ref)
            ref.dfDeleted.connect(self.referencesDeleted)
        self.endInsertRows()
        self.itemsAdded.emit(self.dflist[first:])

    def referencesDeleted(self, ref: DFReference):
        """This method is called with DFReference objects are about to be deleted globally"""
        for i in reversed(range(len(self.dflist))):
            item = self.dflist[i]  # type: DFItem
            if item.ref is ref:
                self.removeRow(i)
                # do not break out of this loop as a plot-list may contain multiple references to the same DF
        self.itemsDeleted.emit()

    def deleteSelectedRows(self, indexes: list):
        selected = sorted(indexes, key=lambda index: index.row())
        rows = [index.row() for index in selected]
        differences = [rows[i + 1] - rows[i] for i in range(len(rows) - 1)]
        sections = list()
        sections.append([rows[0]])
        for diff, row in zip(differences, rows[1:]):
            if diff == 1:  # rows are contiguous:
                sections[-1].append(row)
            else:  # rows are not contiguous
                sections.append([row])
        for section in sections[::-1]:
            self.removeRows(section[0], len(section))
        self.itemsDeleted.emit()


class DFListView(QtWidgets.QTreeView):
    modelClass = DFListModel

    def __init__(self, parent=None):
        super(DFListView, self).__init__(parent)
        self.setSelectionMode(self.ExtendedSelection)
        self.setDragDropMode(self.DragDrop)
        self.setAcceptDrops(True)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.rightClicked)
        self.supportedClasses = (StructuredDataFrame,)  # support all StructuredDataFrame classes by default

        # normally this will be instantiated with no data, so let's set it up, but can be overridden by the user
        self.setRootIsDecorated(False)
        self.initializeModel()
        self.setHeaders(["Datasets"])

    def initializeModel(self):
        self.setModel(
            self.modelClass(
                DFItemList()
            )
        )

    def model(self) -> DFListModel:
        return super(DFListView, self).model()

    def setColumnCount(self, val):
        self.model().setColumnCount(val)

    def setHeaders(self, headers: list):
        self.model().setHeaders(headers)

    def setItemClass(self, cls):
        """
        Set the List-item class that holds the plotting data classes necessary for visualization

        Parameters
        ----------
        cls : type
        """
        self.model().dflist.itemClass = cls

    def setModel(self, model: DFListModel):
        if not isinstance(model, DFListModel):
            raise TypeError("This View only supports models of type {:}".format(DFListModel.__name__))
        super(DFListView, self).setModel(model)

    def isSupported(self, ref: DFReference):
        supported = isinstance(ref.df, self.supportedClasses)
        return supported

    def addDataFrames(self, *refs: DFReference):
        for ref in refs:
            cls = ref.df.__class__
            if not issubclass(cls, self.supportedClasses):
                raise DFTypeError("This List does not support classes of type: {:}".format(cls))
        self.model().addDataFrames(*refs)

    def iterItems(self):
        """
        provide a convenient API to iterate over all the items presented in this view

        Yields
        -------
        DFItem

        """
        for item in self.model().dflist:  # type DFItem
            yield item

    def deleteSelected(self):
        self.model().deleteSelectedRows(self.selectionModel().selectedRows(0))

    def rightClicked(self, pos: QtCore.QPoint):
        menu = QtWidgets.QMenu()

        selected_rows = self.selectionModel().selectedRows(column=0)
        num_dataframes = len(selected_rows)

        if num_dataframes > 0:
            if num_dataframes == 1:
                deleteItem = QtWidgets.QAction("Delete StructuredDataFrame")
            else:
                deleteItem = QtWidgets.QAction("Delete DataFrames")
            deleteItem.triggered.connect(self.deleteSelected)
            menu.addAction(deleteItem)

        menu.exec_(self.mapToGlobal(pos))  # QtWidgets.QAction

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls:
            event.acceptProposedAction()
        else:
            super(DFListView, self).dragEnterEvent(event)

    def row(self, item):
        """

        Parameters
        ----------
        item : DFItem

        Returns
        -------
        int
        """
        return self.model().dflist.index(item)

    def dropEvent(self, event: QtGui.QDropEvent):
        source = event.source()  # type: QtWidgets.QAbstractItemView
        if source is self:  # internal move operation
            super(DFListView, self).dropEvent(event)
            pos = event.pos()
            dropIndex = source.indexAt(pos)
            source.model().internalMove(dropIndex, source.selectionModel().selectedRows(0))
        else:
            if isinstance(source, QtWidgets.QTreeView):
                invalid_drops = []
                valid_drops = []
                for index in source.selectionModel().selectedRows(column=0):  # type: QtCore.QModelIndex
                    pointer = index.internalPointer()  # type: DFReference
                    if self.isSupported(pointer):
                        valid_drops.append(pointer)
                    elif isinstance(pointer, DFReference):
                        invalid_drops.append(pointer.df.__class__.__name__)
                    else:
                        continue

                if valid_drops:
                    self.addDataFrames(*valid_drops)

                if invalid_drops:
                    fn.error_popup(
                        "Invalid StructuredDataFrame types: \n\t" + "\n\t".join(invalid_drops) +
                        "\n\nOnly DataFrames of the following types are allowed:\n\t" +
                        "\n\t".join(cls.__name__ for cls in self.supportedClasses)
                    )


class DFAccessorDelegate(QtWidgets.QItemDelegate):
    def __init__(self, parent=None):
        super(DFAccessorDelegate, self).__init__(parent)

    def createEditor(self, parent: QtWidgets.QWidget, option: 'QStyleOptionViewItem', index: QtCore.QModelIndex):
        comboBox = QtWidgets.QComboBox(parent)
        dflist = index.internalPointer()  # type: DFItemList
        item = dflist[index.row()]  # type: DFItem
        comboBox.addItems(list(map(str, item.accessors.keys())))
        comboBox.currentIndexChanged.connect(self.currentIndexChanged)
        return comboBox

    def setEditorData(self, editor: QtWidgets.QComboBox, index: QtCore.QModelIndex):
        editor.blockSignals(True)
        dflist = index.internalPointer()  # type: DFItemList
        item = dflist[index.row()]  # type: DFItem
        column = index.column()
        if column == 1:
            editor.setCurrentIndex(editor.findText(str(item.x_accessor)))
        elif column == 2:
            editor.setCurrentIndex(editor.findText(str(item.y_accessor)))
        elif column == 3:
            editor.setCurrentIndex(editor.findText(str(item.z_accessor)))
        editor.blockSignals(False)

    def setModelData(self, editor: QtWidgets.QComboBox, model: DFListModel, index: QtCore.QModelIndex):
        item = index.internalPointer()[index.row()]
        model.setData(index, item.accessors.getKey(editor.currentIndex()), QtCore.Qt.EditRole)

    @QtCore.pyqtSlot()
    def currentIndexChanged(self):
        self.commitData.emit(self.sender())


class DFSeriesListView(DFListView):
    def __init__(self, parent=None):
        super(DFSeriesListView, self).__init__(parent)
        self.setHeaders(["Name", "Series"])
        self.setItemDelegateForColumn(1, DFAccessorDelegate(self))


class DFXYListView(DFListView):
    def __init__(self, parent=None):
        super(DFXYListView, self).__init__(parent)
        self.setHeaders(["Label", "x-data", "y-data"])
        self.setItemDelegateForColumn(1, DFAccessorDelegate(self))
        self.setItemDelegateForColumn(2, DFAccessorDelegate(self))


def test(tree=False):
    """
    Instantiate a DFTreeView, add in some data and test the drag and drop
    features.
    """
    from ..plugins.examples import example_powderdiffraction, example_vsm
    from .masterdftree import DFTreeView
    import sys
    df1 = example_powderdiffraction(); df1.name = "Ferrite 1"
    df2 = example_powderdiffraction(); df2.name = "Ferrite 2"
    df3 = example_powderdiffraction(); df3.name = "Ferrite 3"
    df4 = example_vsm(); df4.name = "VSM 1"
    df5 = example_vsm(); df5.name = "VSM 2"
    app = QtWidgets.QApplication([])

    def excepthook_(val, type, tb):
        return sys.__excepthook__(val, type, tb)
    sys.excepthook = excepthook_

    widget = QtWidgets.QWidget()
    hlay = QtWidgets.QHBoxLayout(widget)

    tv = DFTreeView()
    tv.addDataFrame(df1)
    tv.addDataFrame(df2)
    tv.addDataFrame(df3)
    tv.addDataFrame(df4)
    tv.addDataFrame(df5)

    if not tree:
        lv = DFListView()
    else:
        lv = DFXYListView()
    lv.setItemClass(DFItem)  # cannot add items to the list before calling this
    hlay.addWidget(tv)
    hlay.addWidget(lv)

    widget.show()
    sys.exit(app.exec_())



import re
from datetime import datetime
import time


def excel_sheet_name(name, names_list=[], length=31):
    pattern = "[\\/\*\[\]:\?%&]+"
    clean_name = re.sub(pattern, "", name)[:length]

    if clean_name in names_list:
        base_name = clean_name[:length-3]

        for i in range(100):
            # not going to handle i > 99 errors, because I assume we'll never get that high
            clean_name = base_name + "-{:02d}".format(i)
            if clean_name not in names_list:
                break

    return clean_name


def iso_date_string(dtime=None):
    if type(dtime) in (float, int):
        dtime = datetime(dtime)
    elif not isinstance(dtime, datetime):
        dtime = datetime.now()
    return dtime.isoformat()



"""a module with some custom QtWidget Classes"""

from PyQt5 import QtWidgets, QtCore, QtGui


class MdiArea(QtWidgets.QMdiArea):

    urlsDropped = QtCore.pyqtSignal(object)

    def __init__(self, parent=None):
        super(MdiArea, self).__init__(parent)
        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls:
            event.acceptProposedAction()
        else:
            super(MdiArea, self).dragEnterEvent(event)

    def dragMoveEvent(self, event):
        super(MdiArea, self).dragMoveEvent(event)

    def dropEvent(self, dropEvent):
        """
        :type dropEvent: QtGui.QDropEvent
        :return:
        """
        mime = dropEvent.mimeData()
        source = dropEvent.source()

        if mime.hasUrls():
            self.urlsDropped.emit(mime)


class TextWarning(QtWidgets.QDialog):
    def __init__(self, text, parent=None):
        super(TextWarning, self).__init__(parent)
        self.resize(400, 300)
        self.gridLayout = QtWidgets.QGridLayout(self)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.textBrowser = QtWidgets.QTextBrowser(self)
        self.verticalLayout.addWidget(self.textBrowser)
        self.buttonBox = QtWidgets.QDialogButtonBox(self)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.verticalLayout.addWidget(self.buttonBox)
        self.gridLayout.addLayout(self.verticalLayout, 0, 0, 1, 1)

        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        if type(text) in (list, tuple):
            text = "\n".join(text)
        self.textBrowser.setText(text)
        self.setWindowTitle("Warning")


class WarningMsgBox(QtWidgets.QDialog):
    def __init__(self, text, title="Warning", parent=None):
        super(WarningMsgBox, self).__init__(parent)
        self.resize(500, 400)
        self.verticalLayout = QtWidgets.QVBoxLayout(self)
        self.verticalLayout.setContentsMargins(0, 0, 0, -1)
        self.textBrowser = QtWidgets.QTextBrowser(self)
        self.verticalLayout.addWidget(self.textBrowser)
        self.buttonBox = QtWidgets.QDialogButtonBox(self)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Ok)
        self.verticalLayout.addWidget(self.buttonBox)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.setWindowTitle(title)

        if type(text) in (list, tuple):
            text = "\n".join(text)
        self.textBrowser.setText(text)


class Spoiler(QtWidgets.QWidget):
    def __init__(self, parent=None, title='', animationDuration=300):
        """
        References:
            # Adapted from c++ version
            http://stackoverflow.com/questions/32476006/how-to-make-an-expandable-collapsable-section-widget-in-qt
        """
        super(Spoiler, self).__init__(parent=parent)

        self.animationDuration = 300
        self.toggleAnimation = QtCore.QParallelAnimationGroup()
        self.contentArea = QtWidgets.QScrollArea()
        self.headerLine = QtWidgets.QFrame()
        self.toggleButton = QtWidgets.QToolButton()
        self.mainLayout = QtWidgets.QGridLayout()

        toggleButton = self.toggleButton
        toggleButton.setStyleSheet("QToolButton { border: none; }")
        toggleButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        toggleButton.setArrowType(QtCore.Qt.RightArrow)
        toggleButton.setText(str(title))
        toggleButton.setCheckable(True)
        toggleButton.setChecked(False)

        headerLine = self.headerLine
        headerLine.setFrameShape(QtWidgets.QFrame.HLine)
        headerLine.setFrameShadow(QtWidgets.QFrame.Sunken)
        headerLine.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Maximum)

        self.contentArea.setStyleSheet("QScrollArea { background-color: white; border: none; }")
        self.contentArea.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        # start out collapsed
        self.contentArea.setMaximumHeight(0)
        self.contentArea.setMinimumHeight(0)
        # let the entire widget grow and shrink with its content
        toggleAnimation = self.toggleAnimation
        toggleAnimation.addAnimation(QtCore.QPropertyAnimation(self, "minimumHeight"))
        toggleAnimation.addAnimation(QtCore.QPropertyAnimation(self, "maximumHeight"))
        toggleAnimation.addAnimation(QtCore.QPropertyAnimation(self.contentArea, "maximumHeight"))
        # don't waste space
        mainLayout = self.mainLayout
        mainLayout.setVerticalSpacing(0)
        mainLayout.setContentsMargins(0, 0, 0, 0)
        row = 0
        mainLayout.addWidget(self.toggleButton, row, 0, 1, 1, QtCore.Qt.AlignLeft)
        mainLayout.addWidget(self.headerLine, row, 2, 1, 1)
        row += 1
        mainLayout.addWidget(self.contentArea, row, 0, 1, 3)
        self.setLayout(self.mainLayout)

        def start_animation(checked):
            arrow_type = QtCore.Qt.DownArrow if checked else QtCore.Qt.RightArrow
            direction = QtCore.QAbstractAnimation.Forward if checked else QtCore.QAbstractAnimation.Backward
            toggleButton.setArrowType(arrow_type)
            self.toggleAnimation.setDirection(direction)
            self.toggleAnimation.start()

        self.toggleButton.clicked.connect(start_animation)

    def setContentLayout(self, contentLayout):
        # Not sure if this is equivalent to self.contentArea.destroy()
        self.contentArea.destroy()
        self.contentArea.setLayout(contentLayout)
        collapsedHeight = self.sizeHint().height() - self.contentArea.maximumHeight()
        contentHeight = contentLayout.sizeHint().height()
        for i in range(self.toggleAnimation.animationCount()-1):
            spoilerAnimation = self.toggleAnimation.animationAt(i)
            spoilerAnimation.setDuration(self.animationDuration)
            spoilerAnimation.setStartValue(collapsedHeight)
            spoilerAnimation.setEndValue(collapsedHeight + contentHeight)
        contentAnimation = self.toggleAnimation.animationAt(self.toggleAnimation.animationCount() - 1)
        contentAnimation.setDuration(self.animationDuration)
        contentAnimation.setStartValue(0)
        contentAnimation.setEndValue(contentHeight)


class LineEdit(QtWidgets.QLineEdit):
    def __init__(self, heightx=None, widthx=None, parent=None):
        super(LineEdit, self).__init__(parent)
        self.heightx = 1 if heightx is None else heightx
        self.widthx = 1 if widthx is None else widthx
        self.size = super(LineEdit, self).sizeHint()
        self.size.setHeight(self.size.height() * self.heightx)
        self.size.setWidth(self.size.width() * self.widthx)

    def sizeHint(self):
        return self.size


class LineEditNumeric(LineEdit):
    def __init__(self, parent=None):
        super(LineEditNumeric, self).__init__(widthx=0.5, parent=parent)
        self.setValidator(QtGui.QDoubleValidator())

    def value(self):
        return float(self.text())

    def setValue(self, val):
        self.setText("{:0.08g}".format(val))



import sys
import os

from PyQt5 import QtWidgets, QtGui
import pyqtgraph as pg

dpi = None
dpi_scaling = 1.0
dpi_100percent = 96


def set_dpi_scaling():
    app = QtWidgets.QApplication.instance()
    screen = app.screens()[0]  # type QtGui.QScreen
    dpi = screen.logicalDotsPerInch()
    global dpi_scaling
    dpi_scaling = dpi / dpi_100percent


preferred_styles = ['plastique', 'Fusion', 'cleanlooks', 'motif', 'cde']
preferred_style = 'Fusion'

pg.setConfigOption('background', 'w')
pg.setConfigOption('foreground', 'k')
pg.setConfigOption('antialias', True)

resource_manager = None

user_path = os.path.expanduser("~")
executable_dir = os.path.dirname(sys.executable)

qt_dir = os.path.dirname(os.path.abspath(__file__))
radie_dir = os.path.join(qt_dir, "..")
resources_dir = os.path.join(qt_dir, "resources")
icon_path = os.path.join(resources_dir, "icons")

radie_icon = os.path.join(icon_path, "radie.svg")
